--==== set upvalues ====--
-- GLOBALS: CUSTOM_CLASS_COLORS, RAID_CLASS_COLORS, LFG_RETURN_VALUES, MAX_PLAYER_LEVEL
local _G, LibStub, CreateFont, GetExpansionLevel, GetMaxLevelForExpansionLevel, GetLocale, cdt, cmf, cmp, crep, cwm = _G, LibStub, CreateFont, GetExpansionLevel, GetMaxLevelForExpansionLevel, GetLocale, C_DateAndTime, C_MajorFactions, C_Map, C_Reputation, C_UIWidgetManager
local GetRealmName,  UnitClass, UnitLevel, UnitName = GetRealmName,  UnitClass, UnitLevel, UnitName
local GetCurrencyInfo, GetItemCount, GetLogIndexForQuestID, GetQuestObjectives, IsQuestFlaggedCompleted, RequestRaidInfo, RequestLFDPlayerLockInfo, GetNumSavedInstances, GetSavedInstanceEncounterInfo, GetSavedInstanceInfo, GetNumSavedWorldBosses, GetSavedWorldBossInfo = C_CurrencyInfo.GetCurrencyInfo, GetItemCount, C_QuestLog.GetLogIndexForQuestID, C_QuestLog.GetQuestObjectives, C_QuestLog.IsQuestFlaggedCompleted, RequestRaidInfo, RequestLFDPlayerLockInfo, GetNumSavedInstances, GetSavedInstanceEncounterInfo, GetSavedInstanceInfo, GetNumSavedWorldBosses, GetSavedWorldBossInfo
local EJ_GetInstanceInfo, GetLFGDungeonInfo, GetLFGDungeonNumEncounters, GetLFGDungeonRewards, GetRandomDungeonBestChoice = EJ_GetInstanceInfo, GetLFGDungeonInfo, GetLFGDungeonNumEncounters, GetLFGDungeonRewards, GetRandomDungeonBestChoice
local _, date, ipairs, next ,pairs, select, setmetatable, time, tinsert, tsort, type, wipe = _, date, ipairs, next, pairs, select, setmetatable, time, table.insert, table.sort, type, wipe
local strfind, strformat, strgsub, strlower, strmatch, strsub, tostring = string.find, string.format, string.gsub, string.lower, strmatch, string.sub, tostring
local floor, fmod, hugeNumber, max, min = math.floor, math.fmod, math.huge, math.max, math.min
local strNormal, strHeroic, strRaidFinder, strChallenge, strMythic, strLegacy, strWorldBoss, strTimewalker, strUnknown, strInstanceSaved = PLAYER_DIFFICULTY1, PLAYER_DIFFICULTY2, PLAYER_DIFFICULTY3, PLAYER_DIFFICULTY5, PLAYER_DIFFICULTY6, LFG_LIST_LEGACY, RAID_INFO_WORLD_BOSS, PLAYER_DIFFICULTY_TIMEWALKER, UNKNOWN, INSTANCE_SAVED


-- LibDataBroker
local ldb = LibStub:GetLibrary("LibDataBroker-1.1", true)
local LDBIcon = ldb and LibStub("LibDBIcon-1.0",true)
local LibQTip = LibStub('LibQTip-1.0')
local Ailo = LibStub("AceAddon-3.0"):NewAddon("Ailo", "AceConsole-3.0", "AceEvent-3.0", "AceHook-3.0", "AceTimer-3.0")

local L = LibStub("AceLocale-3.0"):GetLocale("Ailo", false)
local DB_VERSION = 35
local maxMythicPlus = 20
local typeMythicPlus = Enum.WeeklyRewardChestThresholdType.Activities
local charLevel
local currentRealmChar
local strfDBKey  -- format string for creating keys for the current character, ie Stormrage.Bob.%s
local CellFormat = {}
local CellFonts = {}

local icons = {
	LFR = "Interface\\LFGFrame\\UI-LFR-PORTRAIT",
	WorldBosses = "Interface\\Icons\\inv_misc_map02",
	[39287] = "136063",  -- Deathtalon -- spell_nature_natureswrath
	[39288] = "132356",  -- Terrorfist -- ability_warrior_secondwind
	[39289] = "252185",  -- Doomroller -- ability_vehicle_siegeenginecannon
	[39290] = "136220",  -- Vengeance  -- spell_shadow_summonsuccubus
	[70723] = "1016352", -- Primal Earth Core -- ability_gift_of_earth
	[70752] = "893777",  -- Primal Water Core -- ability_shawaterelemental_reform
	[70753] = "4554434", -- Primal Air Core   -- inv_10_elementalcombinedfoozles_air
	[70754] = "132847",  -- Primal Fire Core  -- inv_elemental_primal_fire
	[73638] = "237283",  -- Garrison Invasion Bronze   -- inv_misc_coin_19
	[73639] = "237282",  -- Garrison Invasion Silver   -- inv_misc_coin_18
	[37640] = "237281",  -- Garrison Invasion Gold     -- inv-misc-coin-17
	[38482] = "2744751", -- Garrison Invasion Platinum -- inv-misc-azsharacoin2
}


-- Used to color mythic LFG search results depending on lockouts
-- Stored in an easy to compare format when iterating over the LFG frame buttons
local LFGSearchResultsLockouts = {}

-- updateRetry variables used by FirstUpdate timer callback
local updateRetryLimit = 10
local updateRetryCount = 0
local updateRetryTimer = nil
local updateJailerTimer = nil


-- Helperfunctions
local function AbbreviateString(s) return strgsub(s, "(%a)[%l%p]*[%s%-]*", "%1") end

-- breaks one of our db keys into its various parts and pairs
local function breakKey(strKey)
	local dot1 = strfind(strKey, '.', 0, true)
	if not dot1 then -- key of foo returns foo, nil, foo
		return strKey, nil, strKey
	end

	local dot2 = strfind(strKey, '.', dot1 + 1, true)
	if not dot2 then -- key of foo.bar returns foo, bar, foo.bar
		return strsub(strKey, 1, dot1-1), strsub(strKey, dot1+1), strKey
	end

	local dot3 = strfind(strKey, '.', dot2 + 1, true)
	if not dot3 then -- a key of x.y.z returns x, y, x.y, z, nil, z
		return strsub(strKey, 1, dot1-1), strsub(strKey, dot1+1, dot2-1), strsub(strKey, 1, dot2-1), strsub(strKey, dot2+1), nil, strsub(strKey, dot2+1)
	end

	-- a key of a.b.c.d returns a, b, a.b, c, d, c.d
	return strsub(strKey, 1, dot1-1), strsub(strKey, dot1+1, dot2-1), strsub(strKey, 1, dot2 - 1), strsub(strKey, dot2+1, dot3-1), strsub(strKey, dot3+1), strsub(strKey, dot2+1)
end

-- table get - safely returns a nested subtable, creating any tables needed along the way
local function tget(t, k, ...)
	if k == nil then return t end
	t[k] = t[k] or {}
	return tget(t[k], ...)
end

function Ailo:Output(...)
	if self.db.profile.showMessages then
		Ailo:Print(...)
	end
end 

-- table for converting the games instance difficulties into our own instance types
local instanceTypes = setmetatable(
	{
		[DifficultyUtil.ID.DungeonNormal] = "NormalDungeon",
		[DifficultyUtil.ID.DungeonHeroic] = "HeroicDungeon",
		[DifficultyUtil.ID.DungeonChallenge] = "ChallengeDungeon",
		[DifficultyUtil.ID.Raid10Normal] = "LegacyRaid",
		[DifficultyUtil.ID.Raid25Normal] = "LegacyRaid",
		[DifficultyUtil.ID.Raid10Heroic] = "LegacyRaid",
		[DifficultyUtil.ID.Raid25Heroic] = "LegacyRaid",
		[DifficultyUtil.ID.RaidLFR] = "RaidFinder",
		[DifficultyUtil.ID.Raid40] = "LegacyRaid",
		[DifficultyUtil.ID.PrimaryRaidNormal] = "NormalRaid",
		[DifficultyUtil.ID.PrimaryRaidHeroic] = "HeroicRaid",
		[DifficultyUtil.ID.PrimaryRaidMythic] = "MythicRaid",
		[DifficultyUtil.ID.PrimaryRaidLFR] = "RaidFinder",
		-- [18] = "Molten Core",
		[DifficultyUtil.ID.DungeonMythic] = "MythicDungeon",
		[DifficultyUtil.ID.DungeonTimewalker] = "TimewalkerDungeon",
		[DifficultyUtil.ID.RaidTimewalker] = "TimewalkerRaid",
	}, {
		__index = function(t,k) return "Unknown" end
	} )

-- Sorting
local sortedInstanceTypes = { -- sorted order of our instance types
	"TimewalkerRaid",
	"MythicRaid",
	"HeroicRaid",
	"NormalRaid",
	"LegacyRaid",
	"RaidFinder",
	"ChallengeDungeon",
	"TimewalkerDungeon",
	"MythicDungeon",
	"HeroicDungeon",
	"NormalDungeon",
	"Unknown",
}

-- instance type => localized name of difficulty
local instanceTypeDifficulty = {
	TimewalkerRaid = strTimewalker,
	MythicRaid = strMythic,
	HeroicRaid = strHeroic,
	NormalRaid = strNormal,
	LegacyRaid = strLegacy,
	RaidFinder = strRaidFinder,
	ChallengeDungeon  = strChallenge,
	TimewalkerDungeon = strTimewalker,
	MythicDungeon = strMythic,
	HeroicDungeon = strHeroic,
	NormalDungeon = strNormal,
	Unknown = strUnknown,
}

local optionGroupNames = {
	     general = _G.GENERAL_SUBHEADER,
	         pvp = _G.PVP,
	dragonflight = _G.EXPANSION_NAME9,
	 shadowlands = _G.EXPANSION_NAME8,
	         bfa = _G.EXPANSION_NAME7,
	      legion = _G.EXPANSION_NAME6,
	         wod = _G.EXPANSION_NAME5,
	         mop = _G.EXPANSION_NAME4,
}

local sortedOptionGroups = {
	"dragonflight",
	"general",
	"pvp",
	"shadowlands",
	"bfa",
--	"legion",
	"wod",
	"mop",
}

local RAID_CLASS_COLORS_FONTS = {}

local function setColor(info, r, g, b, a)
	Ailo.db.profile[info[#info]] = { r = r, g = g, b = b, a = a }
end

local function getColor(info)
	return Ailo.db.profile[info[#info]].r, Ailo.db.profile[info[#info]].g, Ailo.db.profile[info[#info]].b, Ailo.db.profile[info[#info]].a
end

local function GetWeeklyResetDay()
	-- number is index into day of week; 1 == Sunday
	local resets = { 
		enCN = 3, enUS = 3, esMX = 3, ptBR = 3,                                   -- Tuesday 
		deDE = 4, enGB = 4, esES = 4, frFR = 4, itIT = 4, ptPT = 4, ruRU = 4,     -- Wednesday
		enTW = 5, koKR = 5, zhCN = 5, zhTW = 5,                                   -- Thursday
	}

	local locale = GetLocale()
	local reset = resets[locale]

	if not reset then Ailo:Print(locale, "locale is not recognized, please report this error") end
	return reset or 3
end

local function GetNextDailyReset()
	-- when first logging in, we may get a bogus result from GetQuestResetTime
	-- if thats the case, then just return hugeNumber so that nothing gets purged
	-- GetSecondsUntilDailyReset looks to be new version of GetQuestResetTime
	local seconds = cdt.GetSecondsUntilDailyReset()
	if seconds and seconds > 0 then
		return time() + seconds
	else
		return hugeNumber
	end
end

local function GetNextWeeklyReset()
	local seconds = cdt.GetSecondsUntilWeeklyReset()
	if seconds and seconds > 0 then
		return time() + seconds
	else
		return hugeNumber
	end
end

local function GetNextBiweeklyReset()
	-- if we're not getting a valid number for the weekly reset, then we can't
	-- calculate the biweekly. just return hugeNumber so that nothing gets purged
	local nextWeeklyReset = GetNextWeeklyReset()
	if (nextWeeklyReset == hugeNumber) then
		return hugeNumber
	end

	local midWeek = nextWeeklyReset - 302400 -- 3.5 days worth of seconds
	return (time() > midWeek) and nextWeeklyReset or midWeek
end

--=========================================================================--
-- data table definitions
--=========================================================================--

local defaults = {
	profile = {
		savedraid = { r=1, g=0, b=0, a=1 },
		freeraid  = { r=0, g=1, b=0, a=1 },
		useClassColors       = true,
		useCustomClassColors = true,
		showRealmHeaderLines = false,
		showCharacterRealm   = false,
		minimapIcon = {
			hide = false,
			minimapPos = 220,
			radius = 80,
		},
		showMessages = true,
		showAllChars = false,
		showLeft     = false,
		minLevel = GetMaxLevelForExpansionLevel(GetExpansionLevel()),
		show = {
			['*'] = true,
			["5Man"]            = false,
			DailyHeroicScenario = false,
			DailyScenario       = false,
			TBVictory           = false,
			WGVictory           = false,
			Conquest            = false,
			DailyPVP            = false,
			LootCoins           = false,
			Bounties            = false,
			Callings            = false,
			Covenant            = false,
			Dragonscale         = false,
			Maruuk              = false,
			Iskaara             = false,
			Valdrakken          = false,
			TowerKnowledge      = false,
			AnimaReservoir      = false,
		},
		columnName = { -- NOTE - new entries must also be added to the options UI section
			Seasonal  = L["Seasonal"],
			Dailies   = L["Dailies"],
			Weeklies  = L["Weeklies"],
			Renown    = L["Renown"],
			Callings  = L["Callings"],
			Bounties  = L["Bounties"],
			Knowledge = L["Knowledge"],
			PKTrainers       = AbbreviateString(L["Profession Trainers"]),
			PKTraders        = AbbreviateString(L["Faction Traders"]),
			PKOrders         = AbbreviateString(L["Work Orders"]),
			PKTreatises      = AbbreviateString(L["Draconic Treatises"]),
			NZothAssaults    = AbbreviateString(L["N'Zoth Assaults"]),
			WorldBosses      = AbbreviateString(L["World Bosses"]),
			Total            = AbbreviateString(L["Total"]),
			Weekly           = AbbreviateString(L["Weekly"]),
			Conquest         = AbbreviateString(L["Conquest"]),
			MythicPlus       = AbbreviateString(strMythic) .. '+',
			GarrisonInvasion = AbbreviateString(L["Garrison Invasion"]),
			[WEEKDAY_SUNDAY]    = L["ABBREVIATION_FOR_SUNDAY"],
			[WEEKDAY_MONDAY]    = L["ABBREVIATION_FOR_MONDAY"],
			[WEEKDAY_TUESDAY]   = L["ABBREVIATION_FOR_TUESDAY"],
			[WEEKDAY_WEDNESDAY] = L["ABBREVIATION_FOR_WEDNESDAY"],
			[WEEKDAY_THURSDAY]  = L["ABBREVIATION_FOR_THURSDAY"],
			[WEEKDAY_FRIDAY]    = L["ABBREVIATION_FOR_FRIDAY"],
			[WEEKDAY_SATURDAY]  = L["ABBREVIATION_FOR_SATURDAY"],
		},
		difficultyAbbr = { -- NOTE - new entries must also be added to the options UI section
			[strNormal]     = AbbreviateString(strNormal),
			[strHeroic]     = AbbreviateString(strHeroic),
			[strMythic]     = AbbreviateString(strMythic),
			[strTimewalker] = AbbreviateString(strTimewalker),
			[strChallenge]  = AbbreviateString(strChallenge),
		},
		instanceAbbr = {},
	},
	global = {
		weeklyResetDay = 3,     -- TODO - set default based on region
		nextPurge = hugeNumber,
		nextDailyReset    = 0,
		nextBiweeklyReset = 0,
		nextWeeklyReset   = 0,
		roster = {},        -- structure is {  "realm.char" = { class, level, renown = {}, enchanting, tailoring } }  -- enchanting and tailoring will be whatever two skills they actually have
		-- the key for a reset record is realm.char.resetname - we can do slices in code, but the base data should remain flat
		currencies = {
			total  = {},    -- structure is { "realm.char.currency"  = ( count, max } }
			weekly = {},    -- structure is { "realm.char.currency"  = ( count, max } }
			season = {},    -- structure is { "realm.char.currency"  = ( count, earned, key } }  -- key is currency
		},
		dailies    = {},    -- structure is { "realm.char.resetname" = { count, max } OR { done = true } }  -- resetname may be in form of optiongroup.name
		biweeklies = {},    -- structure is { "realm.char.resetname" = { count, max } OR { done = true } }  -- resetname may be in form of optiongroup.name
		weeklies   = {},    -- structure is { "realm.char.resetname" = { count, max } OR { done = true } }  -- resetname may be in form of optiongroup.name
		                    --                  OR for skill quests  = { skillName = # } } -- # is progress remaining or 0 to represent existence is true
		                    --          OR for mythic plus tracking  = { name, level } OR { level } OR { progress = "5/8" }
		bounties   = {
			resets = {},    -- structure is { "realm.char.resetname" = { count, max } OR { done = true } }
			timers = {      -- structure is resetname = resettime
				['*'] = 0 },     -- default value for bounties which have not yet had a time recorded
		},
		callings   = {},    -- structure is { "realm.char.day" = { count, max } OR { done = true } }  -- day is number for day of the week, Sunday is 1
		lockouts   = {},    -- structure is { "realm.char.name.type" = { count, max, resettime } } } }  -- includes RaidFinder instances
		-- TODO - it appears that traditional raid and instance lockouts may reset at different times depending on server
	},
}


--[====[ The Trackers table - aka, where the magic happens

This table defines most of our trackers, everything to the left of Emmissary Bounties, the World Bosses
and Dungeon/Raid Lockouts. In one place we are describing how, and in what order, they appear in the
tooltip, how they appear in our options UI, and the criteria used to collect a characters completion
progress for that tracked objective.


Think of the structure of this table as a tree with BRANCHES > (optional) TWIGS > LEAVES.

The first layer of the table, the BRANCHES, define the "spanners". These are the caps that span across
one or more columns at the top of the tooltip to indicate that the columns beneath it belong together
as a group. The LEAVES describe the columns that disply in the tooltip with completion progress towards
a given tracker.

The various chunks of information that define a tracker are located at different points of the tree
depending on the needs of that particular tracker; that is why we sometimes have an extra TWIG layer
and other times don't. LEAVES will *always* have a key value set, though, as that identifies how to find
the saved data to fill in the cell in the table. If the parent of a LEAF has set a key value, such as
LootCoins or NZothAssaults, then the LEAVES *must* have keys of the form parentKey.subKey such as
LootCoins.weekly or NZothAssaults.Empire.

The chunks that make up a tracker are OPTION, CRITERIA, LABEL, FORMAT

OPTION - can be set in BRANCH or its child (so, LEAF if BRANCH > LEAF or TWIG if BRANCH > TWIG > LEAF)
This represents the Tracker as it appears to the user in its most basic form, a checkbox in the options UI
that allows them to Track X. That information is represented by the these fields:
          key = uniquely identifies a tracker within Ailo, eg "DailyHeroicDungeon"
         icon = (optional) icon id or path, if set it will show up next to the checkbox in the options UI
   optionName = localized string to tell user what is tracked
   optionDesc = localized string with a more thorough description seen in tooltip when option is moused over
  optionGroup = "general", "pvp" or name of expansion like "dragonflight"


CRITERIA - usually set at the LEAF, only current exception is for multi column currencies like LootCoins
A Tracker uses one or more types of criteria to determine if the character has completed the tracked objective.
Some Trackers have special code located elsewhere in the addon to manage the tracking. Most, however, define
their criteria using one or more of the following subtables. NOTE: a node containing criteria *must* also have a
key specified, either sharing the OPTION key if defined in the same node, or with a subkey like Seasonal.Brewfest.
     quest = {
          reset = "w", "b", or "d" - when does quest reset. letters stand for weekly, biweekly and daily
            max = (optional) number representing how many of the quests must be completed for total completion
   showProgress = (optional) if true and questId is in quest log then save progress towards objective completion in record
       checklog = (optional) if true and questId is in quest log, then data will always be saved; ensures column shows even if no progress
        saveIds = (optional) if at least one quest has been completed, the ids of those remaining will be saved
       questIds = table containing questids, eg { 1234 } or { 123, 456, 789 }

  skillQuest = {
          reset = "w", "b", or "d" - when does quest reset. letters stand for weekly, biweekly and daily
            max = (optional) table, keyed by skill name, indicating # of quests needed for completion. missing skills are treated as 1
   showProgress = (optional) if true and questId is in quest log then save progress towards objective completion in record
	 currencies = (optional) table of currency ids; changes to the quantity of these currencies triggers a QuestUpdate
         skills = { table keyed by skill name. eg { alchemy = { max = 2, ids = { 66373, 66374 } } }
                    ids = table whose values are quest idSkill1
                    max = (optional) number for completion. if nil, we assume one and will show that quest's partial progress if possible

     dungeon = {
             id = dungeonId to check for completion
          first = (optional) number representing start of event date
           last = (optional) number represeting end of event date

     currency = {
             id = currencyId to be tracked
        display = "combined", "seperate" or "seasonal" - will the currency be displayed in one column or seperated into two (eg Total and Weekly)
	               seasonal means one column and that its a type tied to particular patch season, where the cap increases weekly,
                   and all points earned, even those already spent, count towards the cap
           item = itemId used to find quantity of "currency" if its actually an item
            Get = (optional) specifies name of replacement function for GetCurrencyInfo should that not give correct values for that currency

      renown = {
             id = MajorFaction ID whose renown level we are tracking


LABEL - all BRANCHES and LEAVES *must* have labels
The spanners and column headers need to display something so the users know what is in the column. This might be
a text string or an icon. There is a precedence order to these fields and we'll use the first one we find. So, if
icon was set in the node as part of an OPTION, it won't be used in the LABEL if abbr or columnName is set.
        abbr = string, not directly set in any of our tracker definitions because of localiztion
  columnName = a key into the profiledb.columnName table for a localized and user customizable string
   encounter = Encounter Journal ID; result will be fed through abbreviation framework and prepended with icon if set
        icon = icon id or path, note that if the label is solely an icon we make cell font bigger for a bigger icon
    torgWing = number id for wing, used to get localized name from game, which is then fed through instance abbreviation framework
    fullname = string. BRANCH only, idea is that we could programatically figure out that a spanner is covering enough
               columns that it is wide enough for us to use a longer fullname instead of a shorter abbr or icon. This
               functionality is not implemented yet but it doesn't hurt us to set this field on our BRANCHES anyway.


FORMAT - LEAF only, OPTIONAL
Specifies which format function to use when displaying completion progress for a character.
"Default" is assumed if not set.
     format = (optional) name of a format function in the CellFormat table, eg format = "DualCappedCurrency"
  showEmpty = (optional) column should be shown even when there are no saved records (unless options checkbox to show is false)
--]====]

local Trackers = {
	{	fullname = L["Renown"], columnName = "Renown", children = {
			{	key = "DreamWardens",
				format = "Renown",
				icon = 5244643, -- ui_majorfaction_denizens
				optionName = L["Track Dream Wardens"],
				optionDesc = L["Current level of Renown with the Dream Wardens"],
				optionGroup = "dragonflight",
				renown = { id = 2574 },
			},
			{	key = "Niffin",
				format = "Renown",
				icon = 5140835, -- ui_majorfaction_niffen
				optionName = L["Track Niffen"],
				optionDesc = L["Current level of Renown with the Loamm Niffen"],
				optionGroup = "dragonflight",
				renown = { id = 2564 },
			},
			{	key = "Dragonscale",
				format = "Renown",
				icon = 4687628, -- ui_majorfaction_expedition
				optionName = L["Track Dragonscale"],
				optionDesc = L["Current level of Renown with the Dragonscale Expedition"],
				optionGroup = "dragonflight",
				renown = { id = 2507 },
			},
			{	key = "Maruuk",
				format = "Renown",
				icon = 4687627, -- ui_majorfaction_centaur
				optionName = L["Track Maruuk"],
				optionDesc = L["Current level of Renown with the Maruuk Centaur"],
				optionGroup = "dragonflight",
				renown = { id = 2503 },
			},
			{	key = "Iskaara",
				format = "Renown",
				icon = 4687629, -- ui_majorfaction_tuskarr
				optionName = L["Track Iskaara"],
				optionDesc = L["Current level of Renown with the Iskaara Tuskarr"],
				optionGroup = "dragonflight",
				renown = { id = 2511 },
			},
			{	key = "Valdrakken",
				format = "Renown",
				icon = 4687630, -- ui_majorfaction_valdrakken
				optionName = L["Track Valdrakken"],
				optionDesc = L["Current level of Renown with the Valdrakken Accord"],
				optionGroup = "dragonflight",
				renown = { id = 2510 },
			},
			{	key = "Covenant",
				format = "Renown",
				icon = 3726261,
				optionName = L["Track Covenant"],
				optionDesc = L["Current level of Renown with Covenant"],
				optionGroup = "shadowlands",
			}, 	-- criteria = unique tracker type
	}},

	{	fullname = L["Profession Knowledge"], columnName = "Knowledge", children = {
		{	key = "PKTreatises",
			columnName = "PKTreatises",
			optionName = L["Draconic Treatises"],
			optionDesc = L["If the character completed the weekly quests to gain knowledge by reading Draconic Treatises"],
			optionGroup = "dragonflight",
			format = "ProfessionKnowledge",
			skillQuest = { reset = "w", currencies = { 2023, 2024, 2025, 2026, 2027, 2028, 2029, 2030, 2033, 2034, 2035 },
			       skills = {
				       alchemy = { ids = { 74108 }, },
				 blacksmithing = { ids = { 74109 }, },
				    enchanting = { ids = { 74110 }, },
				   engineering = { ids = { 74111 }, },
				     herbalism = { ids = { 74107 }, },
				   inscription = { ids = { 74105 }, },
				 jewelcrafting = { ids = { 74112 }, },
				leatherworking = { ids = { 74113 }, },
				        mining = { ids = { 74106 }, },
				      skinning = { ids = { 74114 }, },
				     tailoring = { ids = { 74115 }, },
		} }, },
		{	key = "PKOrders",
			columnName = "PKOrders",
			optionName = L["Work Orders"],
			optionDesc = L["If the character completed the weekly quests to gain knowledge by filling work orders"],
			optionGroup = "dragonflight",
			format = "ProfessionKnowledge",
			skillQuest = { reset = "w", showProgress = true, skills = {
				       alchemy = { ids = {}, max = 0 },
				 blacksmithing = { ids = { 70589 }, },
				    enchanting = { ids = {}, max = 0 },
				   engineering = { ids = { 70591 }, },
				     herbalism = { ids = {}, max = 0 },
				   inscription = { ids = { 70592 }, },
				 jewelcrafting = { ids = { 70593 }, },
				leatherworking = { ids = { 70594 }, },
				        mining = { ids = {}, max = 0 },
				      skinning = { ids = {}, max = 0 },
				     tailoring = { ids = { 70595 }, },
		} }, },
		{	key = "PKTrainers",
			columnName = "PKTrainers",
			optionName = L["Profession Trainers"],
			optionDesc = L["If the character completed the weekly quests to gain knowledge given by their profession trainers"],
			optionGroup = "dragonflight",
			format = "ProfessionKnowledge",
			skillQuest = { reset = "w", skills = {
				       alchemy = { ids = { 70530, 70531, 70532, 70533, }, }, -- Conflago
				 blacksmithing = { ids = { 70211, 70233, 70234, 70235, }, }, -- Kuroko
				    enchanting = { ids = { 72155, 72172, 72173, 72175, }, }, -- Soragosa
				   engineering = { ids = { 70539, 70540, 70545, 70557, }, }, -- Clinkyclick
				     herbalism = { ids = { 70613, 70614, 70615, 70616, }, }, -- Agrikus
				   inscription = { ids = { 70558, 70559, 70560, 70561, }, }, -- Talendara
				 jewelcrafting = { ids = { 70562, 70563, 70564, 70565, }, }, -- Tuluradormi
				leatherworking = { ids = { 70567, 70568, 70569, 70571, }, }, -- Koruz
				        mining = { ids = { 70617, 70618, 72156, 72157, }, }, -- Sekita
				      skinning = { ids = { 70619, 70620, 72158, 72159, }, }, -- Ralathor
				     tailoring = { ids = { 70572, 70582, 70586, 70587, }, }, -- Fulafong
		} }, },
		{	key = "PKTraders",
			columnName = "PKTraders",
			optionName = L["Faction Traders"],
			optionDesc = L["If the character completed the weekly quests to gain knowledge given by the faction trade representatives as well as weekly collection of Dreambloom, Iridescent Ore and Curious Hide"],
			optionGroup = "dragonflight",
			format = "ProfessionKnowledge",
			skillQuest = { reset = "w", skills = {
				       alchemy = { ids = { 66937, 66938, 66940, 72427, 75363, 75371, 77932, 77933 }, },
				 blacksmithing = { ids = { 66517, 66897, 66941, 72398, 75148, 75569, 77935, 77936 }, },
				    enchanting = { ids = { 66884, 66900, 66935, 72423, 75150, 75865, 77910, 77937 }, },
				   engineering = { ids = { 66890, 66891, 66942, 72396, 75575, 75608, 77891, 77938 }, },
				     herbalism = { ids = { 71857, 71858, 71859, 71860, 71861, 71864 }, max = 6 }, -- collection of Dreambloom
				   inscription = { ids = { 66943, 66944, 66945, 72438, 75149, 75573, 77889, 77914 }, },
				 jewelcrafting = { ids = { 66516, 66949, 66950, 72428, 75362, 75602, 77892, 77912 }, },
				leatherworking = { ids = { 66363, 66364, 66951, 72407, 75354, 75368, 77945, 77946 }, },
				        mining = { ids = { 72160, 72161, 72162, 72163, 72164, 72165 }, max = 6 }, -- collection of Iridescent Ore
				      skinning = { ids = { 70381, 70383, 70384, 70385, 70386, 70389 }, max = 6 }, -- collection of Curious Hide
				     tailoring = { ids = { 66899, 66952, 66953, 72410, 75407, 75600, 77947, 77949 }, },
		} }, },
		{	key = "PKContainers",
			icon = 1060570, -- sturdy shovel icon
			optionName = L["Containers"],
			optionDesc = L["If the character completed the weekly quests to gain knowledge from items found in containers like Disturbed Dirt Piles or Expedition Scout's Packs"],
			optionGroup = "dragonflight",
			format = "ProfessionKnowledge",
			skillQuest = { reset = "w", skills = {
				       alchemy = { max = 2, ids = { 66373, 66374 } }, -- "Experimental Substance",   "Reawakened Catalyst"
				 blacksmithing = { max = 2, ids = { 66381, 66382 } }, -- "Valdrakken Weapon Chain",  "Draconium Blade Sharpener"
				    enchanting = { max = 2, ids = { 66377, 66378 } }, -- "Prismatic Focusing Shard", "Primal Dust"
				   engineering = { max = 2, ids = { 66379, 66380 } }, -- "Eroded Titan Gizmo",       "Watcher Power Core"
				     herbalism = { max = 0, ids = {} },
				   inscription = { max = 2, ids = { 66375, 66376 } }, -- "Phoenix Feather Quill",    "Iskaaran Trading Ledger"
				 jewelcrafting = { max = 2, ids = { 66388, 66389 } }, -- "Ancient Gem Fragments",    "Chipped Tyrstone"
				leatherworking = { max = 2, ids = { 66384, 66385 } }, -- "Mottled Dragon Scales",    "Preserved Animal Parts"
				        mining = { max = 0, ids = {} },
				      skinning = { max = 0, ids = {} },
				     tailoring = { max = 2, ids = { 66386, 66387 } }, -- "Umbral Bone Needle",       "Primalweave Spindle"
		} }, },
		{	key = "PKDrops",
			icon = 236363,  -- 603532, -- a skull
			optionName = L["Drops"],
			optionDesc = L["If the character completed the weekly quests to gain knowledge from items which drop from mobs"],
			optionGroup = "dragonflight",
			format = "ProfessionKnowledge",
			skillQuest = { reset = "w", skills = {
				       alchemy = { max = 2, ids = { 70504, 70511 } }, -- "Decaying Phlegm",         "Elementous Splinter"
				 blacksmithing = { max = 2, ids = { 70512, 70513 } }, -- "Primeval Earth Fragment", "Molten Globule"
				    enchanting = { max = 2, ids = { 70514, 70515 } }, -- "Primalist Aether",        "Primalist Charm"
				   engineering = { max = 2, ids = { 70516, 70517 } }, -- "Keeper's Mark",           "Infinitely Attachable Pair o' Docks"
				     herbalism = { max = 0, ids = {} },
				   inscription = { max = 2, ids = { 70518, 70519 } }, -- "Curious Djardin Rune",    "Draconic Glamour"
				 jewelcrafting = { max = 2, ids = { 70520, 70521 } }, -- "Incandescent Curio",      "Elegantly Engraved Embellishment"
				leatherworking = { max = 2, ids = { 70522, 70523 } }, -- "Ossified Hide",           "Exceedingly Soft Skin"
				        mining = { max = 0, ids = {} },
				      skinning = { max = 0, ids = {} },
				     tailoring = { max = 2, ids = { 70524, 70525 } }, -- "Ohn'aran Weave",          "Stupidly Effective Stitchery"
		} }, },
	}},

	{	fullname = L["Seasonal"], columnName = "Seasonal", children = {
		{	key = "Valor",
			icon = 463447,
			optionName = L["Track Valor Points"],
			optionDesc = L["TRACK_VALOR_POINTS_DESC"],
			optionGroup = "general",
			currency = { id = 1191, display = "seasonal" },
			format = "CurrencySeasonal", },
		{	key = "Conquest",
			icon = 1523630,
			optionName = L["Track Conquest Points"],
			optionDesc = L["TRACK_CONQUEST_POINTS_DESC"],
			optionGroup = "pvp",
			currency = { id = 1602, display = "seasonal" },
			format = "CurrencySeasonal", },
		{	key = "AspectCrests",
			icon = 5062612,
			optionName = L["Aspect's Fragments"],
			optionDesc = L["SHADOWFLAME_CREST_DESC"],
			optionGroup = "dragonflight",
			currency = { id = 2709, display = "seasonal" },
			format = "CurrencySeasonal", },
		{	key = "WyrmCrests",
			icon = 5062642,
			optionName = L["Wyrm's Fragments"],
			optionDesc = L["SHADOWFLAME_CREST_DESC"],
			optionGroup = "dragonflight",
			currency = { id = 2708, display = "seasonal" },
			format = "CurrencySeasonal", },
		{	key = "DrakeCrests",
			icon = 5062624,
			optionName = L["Drake's Fragments"],
			optionDesc = L["SHADOWFLAME_CREST_DESC"],
			optionGroup = "dragonflight",
			currency = { id = 2707, display = "seasonal" },
			format = "CurrencySeasonal", },
		{	key = "WhelplingCrests",
			icon = 5062636,
			optionName = L["Whelplings's Fragments"],
			optionDesc = L["SHADOWFLAME_CREST_DESC"],
			optionGroup = "dragonflight",
			currency = { id = 2706, display = "seasonal" },
			format = "CurrencySeasonal", },
		{	key = "Flightstones",
			icon = 5172976,
			optionName = L["Track Flightstones"],
			optionDesc = L["Track Flightstones"],
			optionGroup = "dragonflight",
			currency = { id = 2245, display = "combined" },
			format = "DualCappedCurrency", },
		{	key = "CatalystCharges",
			icon = 1394953,
			optionName = L["Catalyst Charges"],
			optionDesc = L["Catalyst Charges"],
			optionGroup = "dragonflight",
			currency = { id = 2796, display = "combined" },
			-- quest = { reset = "wc", max = 1, questIds = { 72528 } },
			format = "DualCappedCurrency", },
		{	key = "TowerKnowledge",
			icon = 4037125,
			optionName = L["Track Tower Knowledge"],
			optionDesc = L["TRACK_TOWER_KNOWLEDGE_DESC"],
			optionGroup = "shadowlands",
			currency = { id = 1904, display = "seasonal" },
			format = "CurrencySeasonal", },
	}},

	{	fullname = L["Loot Coins"],
		icon = 1416740,
		key = "LootCoins",
		optionName = L["Track Warforged Seals"], -- TODO - update any time the coin ID changes
		optionDesc = L["Show number of Warforged Seals and whether weekly quest has been completed"], -- TODO - update any time the coin ID changes
		optionGroup = "bfa",
		currency = { id = 1580, display = "seperate" }, -- 1580 = Seal of Wartorn Fate
		quest = { reset = "wc", max = 2, questIds = { 52834, 52835, 52837, 52838, 52839, 52840 } },
		children = {
			{ key = "LootCoins.total",  columnName = "Total",  format = "CurrencySingle" },
			{ key = "LootCoins.weekly", columnName = "Weekly", format = "CurrencySingle" },
		},
		-- in case we ever want to offer options to track older coins as well
		--  BfA   - currency = { id = 1580, display = "seperate" }, -- 1580 = Seal of Wartorn Fate
		--  BfA   - quest = { reset = "wc", max = 2, questIds = { 52834, 52835, 52837, 52838, 52839, 52840 } },
		-- Legion - currency = { id = 1273, display = "seperate" }, -- 1273 = Seal of Broken Fate
		-- Legion - quest = { reset = "wc", max = 3, questIds = { 43510, 43892, 43893, 43894, 43895,  43896, 43897, 47851, 47864, 47865 } },
		--  WoD   - currency = { id = 1129, display = "seperate" }, -- 1129 = Seal of Inevitable Fate
		--  WoD   - quest = { reset = "wc", max = 3, questIds = { 36054, 36055, 36056, 36057, 36058, 37452, 37453, 37454, 37455, 37456, 37457, 37458, 37459, 39020, 39021 } },
		-- WoD 2  - currency = { id = 994, display = "combined", icon = 838813, format = "CurrencySingle" }, -- 994 = Seal of Tempered Fate
		--  MoP   - currency = { id = 776, display = "combined", icon = 514015, format = "CurrencySingle" }, -- 776 = Warforged Seals
	},

	{	fullname = L["Mythic Plus"], columnName = "MythicPlus", children = {
		{	key = "MythicPlusKey",
			icon = 525134, -- matches icon of key in your bags
			format = "MythicPlusKey",
			optionName = L["Track M+ Key"],
			optionDesc = L["Show difficulty and location of your mythic plus key"],
			optionGroup = "general",
			-- is a unique tracker type with special code
		},
		{	key = "MythicPlusReward1",
			icon = 236647, -- Achievement_pvp_p_01
			format = "MythicPlusReward",
			optionName = L["Track M+ Reward 1"],
			optionDesc = L["Show mythic level for your first reward slot"],
			optionGroup = "general",
			-- is a unique tracker type with special code
		},
		{	key = "MythicPlusReward2",
			icon = 236648, -- Achievement_pvp_p_02
			format = "MythicPlusReward",
			optionName = L["Track M+ Reward 2"],
			optionDesc = L["Show mythic level for your second reward slot"],
			optionGroup = "general",
			-- is a unique tracker type with special code
		},
		{	key = "MythicPlusReward3",
			icon = 236649, -- Achievement_pvp_p_03
			format = "MythicPlusReward",
			optionName = L["Track M+ Reward 3"],
			optionDesc = L["Show mythic level for your third reward slot"],
			optionGroup = "general",
			-- is a unique tracker type with special code
		},

	}},

	{	fullname = L["Dailies"], columnName = "Dailies", children = {
		{	key = "Seasonal",
			optionName = L["Track Seasonal bosses"],
			optionDesc = L["If the character has received the reward for a seasonal boss"],
			optionGroup = "general",
			children = {
				{ 	key = "Seasonal.Valentines",
					icon = "Interface\\Icons\\inv_valentinesboxofchocolates02",
					dungeon = { id = 288,
					         first = time( { year = 2024, month = 2,  day =  5, hour = 10, min = 0 } ),
					          last = time( { year = 2024, month = 2,  day = 19, hour = 10, min = 0 } ) }, },
				{ 	key = "Seasonal.Midsummer",
					icon = "Interface\\Icons\\inv_misc_bag_17",
					dungeon	= { id = 286,
					         first = time( { year = 2024, month = 6,  day = 21, hour = 10, min = 0 } ),
					          last = time( { year = 2024, month = 7,  day =  5, hour = 10, min = 0 } ) }, },
				{	key = "Seasonal.Brewfest",
					icon = "Interface\\Icons\\inv_cask_02",
					dungeon = { id = 287,
					         first = time( { year = 2024, month =  9, day = 20, hour = 10, min = 0 } ),
					          last = time( { year = 2024, month = 10, day =  6, hour = 10, min = 0 } ) }, },
				{	key = "Seasonal.HallowsEnd",
					icon = "Interface\\Icons\\inv_misc_bag_28_halloween",
					dungeon = { id = 285,
					         first = time( { year = 2024, month = 10, day = 18, hour = 10, min = 0 } ),
					          last = time( { year = 2024, month = 11, day =  1, hour = 11, min = 0 } ) }, },
				{	key = "Seasonal.WVeilGreench",
					icon = "Interface\\Icons\\inv_holiday_christmas_present_01",
					quest = { reset = "d", max = 1, questIds = { 6983, 7043 }, }, },
				{	key = "Seasonal.WVeilGrumpus",
					icon = "Interface\\Icons\\achievement_worldevent_merrymaker",
					quest = { reset = "d", max = 4, questIds = { 39651, 39649, 39668, 39648 } }, },
			},
		},
		{	key = "DailyHeroicDungeon",
			icon = "Interface\\Icons\\INV_Helmet_08",
			optionName = L["Track 'Daily Heroic'"],
			optionDesc = L["TRACK_DAILY_HEROIC_DESC"],
			optionGroup = "general",
			dungeon = {},  -- special case, id is set during Initialization
		},
		{	key = "DailyHeroicScenario",
			icon = "Interface\\Icons\\inv_misc_lockboxghostiron",
			optionName = L["Track 'Daily Heroic Scenario'"],
			optionDesc = L["TRACK_DAILY_HEROIC_SCENARIO_DESC"],
			optionGroup = "mop",
			dungeon = { id = 641 },
		},
		{	key = "DailyScenario",
			icon = "Interface\\Icons\\Icon_Scenarios",
			optionName = L["Track 'Daily Scenario'"],
			optionDesc = L["TRACK_DAILY_SCENARIO_DESC"],
			optionGroup = "mop",
			dungeon = { id = 493 },
		},
		{	key = "DailyPVP",
			icon = "Interface\\PVPFrame\\Icon-Combat",
			optionName = L["Track PvP daily"],
			optionDesc = L["TRACK_DAILY_PVP_DESC"],
			optionGroup = "pvp",
			-- is a unique tracker type with special code
		},
		{	key = "Tillers",
			icon = "Interface\\Icons\\achievement_faction_tillers",
			optionName = L["Track crops"],
			optionDesc = L["If the character has gathered crops off their farm today"],
			optionGroup = "mop",
			-- is a unique tracker type with special code
		},
		{	key = "ScrapMeltdown",
			icon = "Interface\\Icons\\garrison_weaponupgrade",
			-- optionName = ???, we never put in a checkbox for this o.O
			-- optionDesc = ???, we never put in a checkbox for this o.O
			quest = { reset = "d", max = 2, questIds = { 38175, 38188 } },
		},
		{	key = "HellbaneRares",
			icon = "Interface\\Icons\\inv_box_petcarrier_01",
			optionName = L["Track Hellbane rares"],
			optionDesc = L["Track daily kills of |T136063:0|tDeathtalon, |T132356:0|tTerrorfist, |T252185:0|tDoomroller and |T136220:0|tVengeance"],
			optionGroup = "wod",
			format = "RemainingIcons",
			quest = { reset = "d", saveIds = true, max = 4, questIds = { 39287, 39288, 39289, 39290 } },  -- ids for Deathtalon, Terrorfist, Doomroller, Vengeance
		},
		{	key = "JailerEye",
			icon = 895885,
			format = "SimpleCap",
			optionName = L["Eye of the Jailer"],
			optionDesc = L["Show the amount of attention you have attracted from the Jailer in The Maw"],
			optionGroup = "shadowlands",
			-- is a unique tracker type with special code
		},
	}},

	{	fullname = L["Weeklies"], columnName = "Weeklies", children = {
	-- standard weeklies
		{	key = "WeeklyACallToBattle",
			icon = "interface\\icons\\achievement_bg_killflagcarriers_grabflag_capit",
			optionName = L["A Call to Battle"],
			optionDesc = L["If the character has completed the weekly event to win 4 Battleground Matches"],
			optionGroup = "pvp",
			format = "DefaultEmpty",
			quest = { reset = "w", showProgress = true, checklog = true, questIds = { 72723 } },
		},
		{	key = "WeeklyTheArenaCalls",
			icon = "interface\\icons\\achievement_pvp_legion07",
			optionName = L["The Arena Calls"],
			optionDesc = L["If the character has completed the weekly event to win 10 Arena Skirmishes"],
			optionGroup = "pvp",
			format = "DefaultEmpty",
			quest = { reset = "w", showProgress = true, checklog = true, questIds = { 72720 } },
		},
		{	key = "WeeklyTheVeryBest",
			icon = "interface\\icons\\inv_pet_battlepettraining",
			optionName = L["The Very Best"],
			optionDesc = L["If the character has completed the weekly event to win 5 PvP Pet Battles"],
			optionGroup = "general",
			format = "DefaultEmpty",
			quest = { reset = "w", showProgress = true, checklog = true, questIds = { 72721 } },
		},
		{	key = "WeeklyTheWorldAwaits",
			icon = "Interface\\GossipFrame\\AvailableQuestIcon",
			optionName = L["The World Awaits"],
			optionDesc = L["If the character has completed the weekly event to finish 10 world quests"],
			optionGroup = "general",
			format = "DefaultEmpty",
			quest = { reset = "w", showProgress = true, checklog = true, questIds = { 72728 } },
		},
		{	key = "WeeklyAPathThroughTime",
			icon = "interface\\icons\\spell_holy_borrowedtime",
			optionName = L["A Path Through Time"],
			optionDesc = L["If the character has completed the weekly event to defeat 5 Timewalking raid bosses or dungeons"],
			optionGroup = "general",
			format = "DefaultEmpty",
			quest = { reset = "w", showProgress = true, checklog = true, questIds = { 72719, 72724, 72725, 72726, 72727, 72810 } },
		},
		{	key = "WeeklyEmmissaryOfWar",
			icon = "interface\\icons\\inv_bijou_purple",
			optionName = L["Emmissary of War"],
			optionDesc = L["If the character has completed the weekly event to finish 4 Mythic dungeons"],
			optionGroup = "general",
			format = "DefaultEmpty",
			quest = { reset = "w", showProgress = true, checklog = true, questIds = { 72722 } },
		},
		{	key = "WeeklyFightingIsItsOwnReward",
			icon = "interface\\icons\\inv_bijou_blue",
			optionName = L["Fighting is Its Own Reward"],
			optionDesc = L["If the character has completed the weekly quest to finish 5 Heroic dungeons"],
			optionGroup = "general",
			format = "DefaultEmpty",
			quest = { reset = "w", showProgress = true, checklog = true, questIds = { 76122 } },
		},
	-- reputation
		{	key = "WorthyAllyDreamWardens",
			icon = 133669,
			optionName = L["A Worthy Dream Warden"],
			optionDesc = L["If the character completed the A Worthy Ally: Dream Wardens weekly quest"],
			optionGroup = "dragonflight",
			format = "DefaultEmpty",
			quest = { reset = "w", checklog = true, questIds = { 78444 } }, },
		{	key = "WorthyAllyNiffen",
			icon = 133660,
			optionName = L["A Worthy Loamm Niffen"],
			optionDesc = L["If the character completed the A Worthy Ally: Loamm Niffen weekly quest"],
			optionGroup = "dragonflight",
			format = "DefaultEmpty",
			quest = { reset = "w", checklog = true, questIds = { 75665 } }, },
		{	key = "AidingTheAccord",
			icon = 1603189,
			optionName = L["Aiding the Accord"],
			optionDesc = L["If the character completed the Aiding the Accord weekly quest"],
			optionGroup = "dragonflight",
			format = "DefaultEmpty",
			quest = { reset = "w", checklog = true, questIds = { 70750, 72068, 72373, 72374, 72375, 75259, 75859, 75860, 75861, 77254, 77976, 78446, 78447 } }, },
	-- regularly occuring events
		{	key = "Superbloom",
			icon = 134210, -- icon for the Cache of Overblooming Treasures
			optionName = L["The Superbloom"],
			optionDesc = L["If the character completed The Superbloom quest this week"],
			optionGroup = "dragonflight",
			showEmpty = true,
			quest = { reset = "w", max = 1, questIds = { 78319 } }, },
		{	key = "Suffusion Camp",
			icon = 134153,
			optionName = L["Suffusion Camp"],
			optionDesc = L["If the character has finished the Suffusion Camp world quest and killed Fyrakk's Disciple"],
			optionGroup = "dragonflight",
			showEmpty = true,
			quest = { reset = "w", max = 2, questIds = { 74501, 75280, 75467 } }, },
		{	key = "ResearchersUnderFire",
			icon = 136186,
			optionName = L["Researchers Under Fire"],
			optionDesc = L["If the character has collected a Researchers Gift, the Tier level of the reward will be shown. If the Tier 4 reward has been collected we will show it as completed."],
			optionGroup = "dragonflight",
			showEmpty = true,
			format = "ResearchersUnderFire",
			quest = { reset = "w", saveIds = true, questIds = { 75627, 75628, 75629, 75630 } }, },
		{	key = "StormsFury",
			icon = 463562,
			optionName = L["Storm's Fury"],
			optionDesc = L["If the character completed the The Storm's Fury event in the Primalist Future"],
			optionGroup = "dragonflight",
			format = "ChestQuality",
			showEmpty = true,
			quest = { reset = "w", max = 1, questIds = { 73162 } }, },
			-- ???? represents blue chests. flagged and then unflagged when the next siege happens, doesn't degrade to green/white like hunts
			-- 74378 is turned in but not flagged completed when finishing the event
			-- useful as a trigger to know to check for actual completion
		{	key = "SiegeOfDragonbaneKeep",
			icon = 236469,
			optionName = L["Siege of Dragonbane Keep"],
			optionDesc = L["If the character completed the Siege of Dragonebane Keep event"],
			optionGroup = "dragonflight",
			format = "ChestQuality",
			showEmpty = true,
			quest = { reset = "w", max = 1, questIds = { 70866 } }, },
			-- 74295 represents blue chests. flagged and then unflagged when the next siege happens, doesn't degrade to green/white like hunts
			-- 70867 is turned in but not flagged completed when finishing the siege
			-- useful as a trigger to know to check for actual completion
		{	key = "GrandHunts",
			icon = 237378,
			optionName = L["Grand Hunts"],
			optionDesc = L["If the character completed three Grand Hunt events"],
			optionGroup = "dragonflight",
			format = "ChestQuality",
			showEmpty = true,
			quest = { reset = "w", max = 3, questIds = { 70906, 71136, 71137 } }, },
			-- 70907 is turned in but not flagged completed when finishing a hunt
			-- useful as a trigger to know to check for actual completion
		{	key = "CommunityFeast",
			icon = 629056,
			optionName = L["Community Feast"],
			optionDesc = L["If the character completed the Community Feast event"],
			optionGroup = "dragonflight",
			showEmpty = true,
			quest = { reset = "w", max = 1, questIds = { 70893 } }, },
	-- quests with tokens
		{	key = "Blooming Dreamseeds",
			icon = 306845, -- icon for the Seedbloom
			optionName = L["Blooming Dreamseeds"],
			optionDesc = L["If the character acquired a Seedbloom from the Blooming Dreamseeds quest this week"],
			optionGroup = "dragonflight",
			showEmpty = true,
			quest = { reset = "w", max = 1, questIds = { 78821 } }, },
		{	key = "DreamsurgeChrysalis",
			icon = 4643985, -- icon for the dreamsurge chryssalis
			optionName = L["Dreamsurge Chrysalis"],
			optionDesc = L["If the character acquired a Chrysalis from a Dreamsurge this week"],
			optionGroup = "dragonflight",
			showEmpty = true,
			quest = { reset = "w", max = 1, questIds = { 77251 } }, },
		{	key = "DilatedTimeCapsule",
			icon = 2026009, -- icon for the dilated time capsule
			optionName = L["Dilated Time Capsule"],
			optionDesc = L["If the character acquired a Dilated Time Capsule from a Time Rift this week"],
			optionGroup = "dragonflight",
			showEmpty = true,
			quest = { reset = "w", max = 1, questIds = { 77836 } }, },
		{	key = "EncapsulatedDestiny",
			icon = 1391676, -- icon for the encapsulated destiny item
			optionName = L["Encapsulated Destiny"],
			optionDesc = L["If the character acquired an Encapsulated Destiny from Chromie this week"],
			optionGroup = "dragonflight",
			showEmpty = true,
			quest = { reset = "w", max = 1, questIds = { 77172 } }, },
	-- other
		{	key = "Mettle",
			icon = 4643977, -- artisan's mettle
			optionName = L["Show Your Mettle"],
			optionDesc = L["If the character completed the Show Your Mettle quest"],
			optionGroup = "dragonflight",
			showEmpty = true,
			quest = { reset = "w", max = 1, questIds = { 70221 } }, },
		{	key = "PrimalCores",
			icon = 651737,
			optionName = L["Primal Cores"],
			optionDesc = L["If the character has turned in Primal Cores of each element type"],
			optionGroup = "dragonflight",
			showEmpty = true,
			format = "RemainingIcons",
			quest = { reset = "w", saveIds = true, max = 4, questIds = { 70753, 70723, 70754, 70752 } },
		},
		{	key = "TrialOfFlood",
			icon = 135861,
			optionName = L["Trial of Flood"],
			optionDesc = L["If the character completed the Trial of Flood event"],
			optionGroup = "dragonflight",
			showEmpty = true,
			quest = { reset = "w", max = 1, questIds = { 71033 } }, },
		{	key = "TrialOfElements",
			icon = 136028,
			optionName = L["Trial of Elements"],
			optionDesc = L["If the character completed the Trial of Elements event"],
			optionGroup = "dragonflight",
			showEmpty = true,
			quest = { reset = "w", max = 1, questIds = { 71995 } }, },
		{	key = "TrialOfStorms",
			icon = 237589,
			optionName = L["Trial of Storms"],
			optionDesc = L["If the character completed the Trial of Storms event"],
			optionGroup = "dragonflight",
			showEmpty = true,
			quest = { reset = "w", max = 1, questIds = { 74567 } }, },
	--[===[ 10.1 - these aren't being offered, perhaps will come back when S2 starts
		-- the problem is quest completion did not reset for the one I did before patch, so its showing up in grid
		{	key = "ArchivesDungeon",
			-- icon = 463900, -- icon is that of anima reward. commendation icons: Court of Harvesters = 3514227, Undying Army = 3492310, Wild Hunt = 3575389, Ascended = 3641395
			optionName = L["Archival Studies Dungeons"],
			optionDesc = L["If the character has completed the weekly dungeon quests from the Azure Archives Annex"],
			optionGroup = "dragonflight",
			children = {
				{	key = "ArchivesDungeon.66860", encounter = 1197, quest = { reset = "w", checklog = true, questIds = { 66860 } }, }, -- Uldaman: Legacy of Tyr
				{	key = "ArchivesDungeon.66861", encounter = 1199, quest = { reset = "w", checklog = true, questIds = { 66861 } }, }, -- Neltharus
				{	key = "ArchivesDungeon.66862", encounter = 1202, quest = { reset = "w", checklog = true, questIds = { 66862 } }, }, -- Ruby Life Pools
				{	key = "ArchivesDungeon.66863", encounter = 1198, quest = { reset = "w", checklog = true, questIds = { 66863 } }, }, -- The Nokhud Offensive
				{	key = "ArchivesDungeon.66864", encounter = 1196, quest = { reset = "w", checklog = true, questIds = { 66864 } }, }, -- Brackenhide Hollow
				{	key = "ArchivesDungeon.66865", encounter = 1203, quest = { reset = "w", checklog = true, questIds = { 66865 } }, }, -- The Azure Vault
				{	key = "ArchivesDungeon.66866", encounter = 1204, quest = { reset = "w", checklog = true, questIds = { 66866 } }, }, -- Halls of Infusion
				{	key = "ArchivesDungeon.66867", encounter = 1201, quest = { reset = "w", checklog = true, questIds = { 66867 } }, }, -- Algeth'ar Academy
				{	key = "ArchivesDungeon.66868", encounter = 1197, quest = { reset = "w", checklog = true, questIds = { 66868 } }, }, -- Uldaman: Legacy of Tyr
				-- {	key = "ArchivesDungeon.66869", encounter = , quest = { reset = "w", checklog = true, questIds = { 66869 } }, }, -- 
				{	key = "ArchivesDungeon.66870", encounter = 1202, quest = { reset = "w", checklog = true, questIds = { 66870 } }, }, -- Ruby Life Pools
				{	key = "ArchivesDungeon.66871", encounter = 1198, quest = { reset = "w", checklog = true, questIds = { 66871 } }, }, -- The Nokhud Offensive
				-- {	key = "ArchivesDungeon.66872", encounter = , quest = { reset = "w", checklog = true, questIds = { 66872 } }, }, -- 
				{	key = "ArchivesDungeon.66873", encounter = 1203, quest = { reset = "w", checklog = true, questIds = { 66873 } }, }, -- The Azure Vault
				{	key = "ArchivesDungeon.66874", encounter = 1204, quest = { reset = "w", checklog = true, questIds = { 66874 } }, }, -- Halls of Infusion
				{	key = "ArchivesDungeon.66875", encounter = 1201, quest = { reset = "w", checklog = true, questIds = { 66875 } }, }, -- Algeth'ar Academy
			},
		},
	--]===]
		{	key = "BrokerDungeon",
			icon = 463900, -- icon is that of anima reward. commendation icons: Court of Harvesters = 3514227, Undying Army = 3492310, Wild Hunt = 3575389, Ascended = 3641395
			optionName = L["Broker Dungeons"],
			optionDesc = L["If the character has completed the weekly dungeon quests the Brokers in Oribos"],
			optionGroup = "shadowlands",
			children = {
				{	key = "BrokerDungeon.60242", encounter = 1182, icon = 463900, quest = { reset = "w", checklog = true, questIds = { 60242 } }, }, -- Necrotic Wake
				{	key = "BrokerDungeon.60243", encounter = 1189, icon = 463900, quest = { reset = "w", checklog = true, questIds = { 60243 } }, }, -- Sanguine Depths
				{	key = "BrokerDungeon.60244", encounter = 1185, icon = 463900, quest = { reset = "w", checklog = true, questIds = { 60244 } }, }, -- Halls of Atonement
				{	key = "BrokerDungeon.60245", encounter = 1188, icon = 463900, quest = { reset = "w", checklog = true, questIds = { 60245 } }, }, -- De Other Side
				{	key = "BrokerDungeon.60246", encounter = 1184, icon = 463900, quest = { reset = "w", checklog = true, questIds = { 60246 } }, }, -- Mists of Tirna Scithe
				{	key = "BrokerDungeon.60247", encounter = 1187, icon = 463900, quest = { reset = "w", checklog = true, questIds = { 60247 } }, }, -- Theater of Pain
				{	key = "BrokerDungeon.60248", encounter = 1183, icon = 463900, quest = { reset = "w", checklog = true, questIds = { 60248 } }, }, -- Plaguefall
				{	key = "BrokerDungeon.60249", encounter = 1186, icon = 463900, quest = { reset = "w", checklog = true, questIds = { 60249 } }, }, -- Spires of Ascension
				{	key = "BrokerDungeon.60250", encounter = 1187, icon = 463900, quest = { reset = "w", checklog = true, questIds = { 60250 } }, }, -- Theater of Pain
				{	key = "BrokerDungeon.60251", encounter = 1183, icon = 463900, quest = { reset = "w", checklog = true, questIds = { 60251 } }, }, -- Plaguefall
				{	key = "BrokerDungeon.60252", encounter = 1186, icon = 463900, quest = { reset = "w", checklog = true, questIds = { 60252 } }, }, -- Spires of Ascension
				{	key = "BrokerDungeon.60253", encounter = 1182, icon = 463900, quest = { reset = "w", checklog = true, questIds = { 60253 } }, }, -- Necrotic Wake
				{	key = "BrokerDungeon.60254", encounter = 1184, icon = 463900, quest = { reset = "w", checklog = true, questIds = { 60254 } }, }, -- Mists of Tirna Scithe
				{	key = "BrokerDungeon.60255", encounter = 1188, icon = 463900, quest = { reset = "w", checklog = true, questIds = { 60255 } }, }, -- De Other Side
				{	key = "BrokerDungeon.60256", encounter = 1185, icon = 463900, quest = { reset = "w", checklog = true, questIds = { 60256 } }, }, -- Halls of Atonement
				{	key = "BrokerDungeon.60257", encounter = 1189, icon = 463900, quest = { reset = "w", checklog = true, questIds = { 60257 } }, }, -- Sanguine Depths
			},
		},
		{	key = "Torghast",
			icon = 3743738, -- soul ash icon
			optionName = L["Track Torghast"],
			optionDesc = L["Show layers cleared in Torghast wings"],
			optionGroup = "shadowlands",
			children = {
				{	key = "Torghast.SkoldusHall",
					quest = { reset = "w", max = 12, questIds = { 59328, 59329, 59330, 59331, 59332, 59333, 61969, 61970, 63868, 63869, 63870, 63871 } },
					format = "SimpleCap",
					torgWing = 1, },
				{	key = "Torghast.FractureChambers",
					quest = { reset = "w", max = 12, questIds = { 58186, 58187, 58188, 58189, 58190, 58191, 61971, 61972, 63872, 63873, 63874, 63875 } },
					format = "SimpleCap",
					torgWing = 2, },
				{	key = "Torghast.Soulforges",
					quest = { reset = "w", max = 12, questIds = { 58192, 58193, 58194, 58195, 58196, 58197, 61973, 61974, 63876, 63877, 63878, 63879 } },
					format = "SimpleCap",
					torgWing = 3, },
				{	key = "Torghast.ColdheartInterstitia",
					quest = { reset = "w", max = 12, questIds = { 58198, 58199, 58200, 58201, 58202, 58203, 61975, 61976, 63880, 63881, 63882, 63883 } },
					format = "SimpleCap",
					torgWing = 4, },
				{	key = "Torghast.Mortregar",
					quest = { reset = "w", max = 12, questIds = { 58204, 58205, 59326, 59334, 59335, 59336, 61977, 61978, 63884, 63885, 63886, 63887 } },
					format = "SimpleCap",
					torgWing = 5, },
				{	key = "Torghast.UpperReaches",
					quest = { reset = "w", max = 12, questIds = { 59337, 61101, 61131, 61132, 61133, 61134, 61979, 61980, 63888, 63889, 63890, 63891 } },
					format = "SimpleCap",
					torgWing = 6, },
			},
		},
		{	key = "WrathOfTheJailer", -- should we do this as a world boss?
			icon = 135882, -- not correct but it does have crossed swords
			optionName = L["Wrath of the Jailer"],
			optionDesc = L["If the character has completed the Wrath of the Jailer event in The Maw"],
			optionGroup = "shadowlands",
			quest = { reset = "w", max = 1, questIds = { 63414 } }, },
		{	key = "TormentorsOfTorghast", -- should we do this as a world boss?
			icon = 1542854, -- cache - would rather have thing on map
			optionName = L["Tormentors of Torghast"],
			optionDesc = L["If the character has completed the Tormentors of Torghast event in The Maw"],
			optionGroup = "shadowlands",
			quest = { reset = "w", max = 1, questIds = { 63854 } }, },
		{	key = "BeastwarrensHunt",
			icon = 133732, -- not correct, but it has a beast skull
			optionName = L["Beastwarrens Hunt"],
			optionDesc = L["If the character has completed the currently active Hunt in the Beastwarrens"],
			optionGroup = "shadowlands",
			quest = { reset = "b", max = 1, questIds = { 63180, 63194, 63198, 63199 } }, },
		{	key = "ShapingFate",
			icon = "Interface\\Icons\\item_korthia_paragonchest_01", -- 4066373 -- Death's Advance War Chest
			optionName = L["Shaping Fate"],
			optionDesc = L["If the character has subdued teh Jailer's efforts for the week"],
			optionGroup = "shadowlands",
			quest = { reset = "w", max = 1, questIds = { 63949 } }, },
		{	key = "AnimaReservoir",
			icon = 3528288, -- icon from anima currency
			optionName = L["Replenish the Reservoir"],
			optionDesc = L["If the character has finished collecting 1000 anima for their reservoir"],
			optionGroup = "shadowlands",
			quest = { reset = "w", max = 1, questIds = { 61981, 61982, 61983, 61984 } }, },
		{	key = "LostSouls",
			icon = 1391776, -- redeemed soul icon
			optionName = L["Return Lost Souls"],
			optionDesc = L["If the character has finished collecting souls in The Maw"],
			optionGroup = "shadowlands",
			quest = { reset = "w", max = 1, questIds = {
				61331, 62858, 62859, 62860, -- night fae
				61332, 62861, 62862, 62863, -- kyrian
				61333, 62864, 62865, 62866, -- undying army
				61334, 62867, 62868, 62869, -- venthyr
		} }, },
		{	key = "CovenantAssaults",
			optionName = L["Track Covenant Assaults"],
			optionDesc = L["If the character has completed the active covenant assault in The Maw"],
			optionGroup = "shadowlands",
			children = {
				{	key = "CovenantAssaults.NightFae",
					icon = "Interface\\Icons\\ui_sigil_nightfae", -- 3257750,
					quest = { reset = "b", max = 1, questIds = { 63823 } }, },
				{	key = "CovenantAssaults.Necrolord",
					icon = "Interface\\Icons\\ui_sigil_necrolord", -- 3257749,
					quest = { reset = "b", max = 1, questIds = { 63543 } }, },
				{	key = "CovenantAssaults.Venthyr",
					icon = "Interface\\Icons\\ui_sigil_venthyr", -- 3257751,
					quest = { reset = "b", max = 1, questIds = { 63822 } }, },
				{	key = "CovenantAssaults.Kyrian",
					icon = "Interface\\Icons\\ui_sigil_kyrian", -- 3257748,
					quest = { reset = "b", max = 1, questIds = { 63824 } }, },
		}},
		{	key = "Islands",
			icon = "Interface\\Icons\\icon_treasuremap",
			optionName = L["Track Islands"],
			optionDesc = L["If the character has completed their weekly Island Expeditions"],
			optionGroup = "bfa",
			quest = { reset = "w", max = 2, questIds = { 53435, 53436 } }, },
		{	key = "GarrisonInvasion",
			columnName = "GarrisonInvasion",
			optionName = L["Garrison Invasion"],
			optionDesc = L["If the character has defended their Garrison and if any higher level of awards remain"],
			optionGroup = "wod",
			format = "RemainingIcons",
			quest = { reset = "w", max = 4, questIds = { 38482, 37640, 37639, 37638 } }, },
		{	key = "TBVictory",
			icon = 409548,
			optionName = L["Track 'TB Victory'"],
			optionDesc = L["If the character has done the 'Victory in Tol Barad' weekly pvp quest"],
			optionGroup = "pvp",
			quest = { reset = "w", max = 1, questIds = { 28882, 28884 } }, },
		{	key = "WGVictory",
			icon = 133408,
			optionName = L["Track 'WG Victory'"],
			optionDesc = L["If the character has done the 'Victory in Wintergrasp' weekly pvp quest"],
			optionGroup = "pvp",
			quest = { reset = "w", max = 1, questIds = { 13181, 13183 } }, },
	}},
	{	fullname = L["N'Zoth Assaults"],
		columnName = "NZothAssaults",

		key = "NZothAssaults",
		icon = "Interface\\Icons\\inv_legion_cache_nightmare",
		optionName = L["Track N'Zoth Assaults"],
		optionDesc = L["If the character has Rajani and Uldum Accord weekly rewards available"],
		optionGroup = "bfa",
		children = {
			{	key = "NZothAssaults.Empire",
				icon = "Interface\\Icons\\inv_legion_cache_nightmare",  -- 1471396
				quest = { reset = "w", max = 1, questIds = { 56064, 57157 } }, },
			{	key = "NZothAssaults.Amathet",
				icon = "Interface\\Icons\\inv_legion_cache_valajar",  -- 1394950
				quest = { reset = "b", max = 1, questIds = { 55350 } }, },
			{	key = "NZothAssaults.Aqir",
				icon = "Interface\\Icons\\inv_misc_ahnqirajtrinket_03",  -- 133572
				quest = { reset = "b", max = 1, questIds = { 56308 } }, },
			{	key = "NZothAssaults.Mantid",
				icon = "Interface\\Icons\\inv_misc_archaeology_mantidbasket_01",  -- 839401
				quest = { reset = "b", max = 1, questIds = { 57728 } }, },
			{	key = "NZothAssaults.Mogu",
				icon = "Interface\\Icons\\inv_misc_treasurechest03b",  -- 1542852
				quest = { reset = "b", max = 1, questIds = { 57008 } }, },
	}},
}

-- TODO - review getting these into the Trackers table like we did with covenant renown
local SpecialTrackers = {
	{	key = "Callings",
		optionName = L["Track Callings"],
		optionDesc = L["If the character has Covenant Callings available"],
		optionGroup = "shadowlands", },
	{	key = "Bounties",
		optionName = L["Track Emissary Bounties"],
		optionDesc = L["If the character has Emissary Bounties available"],
		optionGroup = "bfa", },
	{	key = "WorldBosses",
		icon = icons.WorldBosses,
		optionName = L["Track World Bosses"],
		optionDesc = L["Tracks any World Bosses from which loot has been collected this week"],
		optionGroup = "general", },
	{	key = "RaidFinder",
		icon = icons.LFR,
		optionName = L["Track Raid Finder instances"],
		optionDesc = L["Tracks any raid done through the Raid Finder tool"],
		optionGroup = "general", },
	{	key = "5Man",
		optionName = L["Show 5-man instances"],
		optionGroup = "general", },
}


-- Every day a random inactive bounty activates with a 3 day expiration timer

local bounty_list_legion = {
	         ["Argussian Reach"] = { icon = 1708505, quest = 48642, },
	    ["Armies of Legionfall"] = { icon = 1708507, quest = 48641, },
	       ["Army of the Light"] = { icon = 1708506, quest = 48639, },
	       ["Court of Farondis"] = { icon = 1394952, quest = 42420, },
	            ["Dreamweavers"] = { icon = 1394953, quest = 42170, },
	      ["Highmountain Tribe"] = { icon = 1394954, quest = 42233, },
	["The Kirin Tor of Dalaran"] = { icon = 1394955, quest = 43179, },
	         ["The Nightfallen"] = { icon = 1394956, quest = 42421, },
	             ["The Wardens"] = { icon = 1394958, quest = 42422, },
	                ["Valarjar"] = { icon = 1394957, quest = 42234, },
}
local bounty_list_bfa = {
	          -- alliance
	          ["7th Legion"] = { icon = 2203912, quest = 50605 },
	     ["Order of Embers"] = { icon = 2203915, quest = 50600 },
	 ["Proudmore Admiralty"] = { icon = 2203916, quest = 50599 },
	        ["Storm's Wake"] = { icon = 2203917, quest = 50601 },
	    ["Waveblade Ankoan"] = { icon = 2909043, quest = 56119 },
	            -- shared
	["Champions of Azeroth"] = { icon = 2203913, quest = 50562 },
	   ["Tortollan Seekers"] = { icon = 2203919, quest = 50604 },
	 ["Rustbolt Resistance"] = { icon = 2909316, quest = 55348 },
	             -- horde
	          ["Honorbound"] = { icon = 2203914, quest = 50606 },
	["Talanji's Expedition"] = { icon = 2203918, quest = 50602 },
	          ["Unshackled"] = { icon = 2909044, quest = 56120 },
	            ["Voldunai"] = { icon = 2203920, quest = 50603 },
	    ["Zandalari Empire"] = { icon = 2203921, quest = 50598 },
}


-- World Boss list
-- Info gets added to TrackedQuests and RepeatableQuests by InitQuestsTable
-- Key is id for EJ_GetEncounterInfo, from which we get the localized name of the boss.
-- Value is the set of quest ids (usually one, but we have option to group mutually exclusive bosses into one entry)
local WorldBosses = {
	-- Legion
	[1790] = { 43512 }, -- Ana-Mouz
	[1774] = { 43193 }, -- Calamir
	[1789] = { 43448 }, -- Drugon the Frostblood
	[1795] = { 43985 }, -- Flotsam
	[1770] = { 42819 }, -- Humongris
	[1769] = { 43192 }, -- Levantus
	[1783] = { 43513 }, -- Na'zak the Fiend
	[1749] = { 42270 }, -- Nithogg
	[1763] = { 42779 }, -- Shar'thos
	[1756] = { 42269 }, -- The Soultakers
	[1796] = { 44287 }, -- Withered J'im
    -- Broken Shore (Legion 7.2)
	[1883] = { 46947 }, -- Brutallus
	[1884] = { 46948 }, -- Malificus
	[1885] = { 46945 }, -- Si'Vash
	[1956] = { 47061 }, -- Apocorn
	-- Argus (Legion 7.3)
	[2010] = { 49169 }, -- Matron Folnuna
	[2011] = { 49167 }, -- Mistress Alluradel
	[2012] = { 49166 }, -- Inquisitor Meto
	[2013] = { 49170 }, -- Occularus
	[2014] = { 49171 }, -- Sotanathor
	[2015] = { 49168 }, -- Pit Lord Vilemus
	-- [2002] = {  }, -- Keeper Aedis
	-- [] = {  }, -- Void-Blade Zedaat
	-- Battle for Azeroth
	[2199] = { 52163 }, -- Azurethos, The Winged Typhoon
	[2210] = { 52196 }, -- Dunegorger Kraulok
	[2197] = { 52157 }, -- Hailstone Construct
	[2141] = { 52169 }, -- Ji'arak
	[2139] = { 52181 }, -- T'zane
	[2198] = { 52166 }, -- Warbringer Yenajz
	-- Stromgarde and Darkshore Warfronts
	[2213] = { 52847 }, -- Doom's Howl
	[2212] = { 52848 }, -- The Lion's Roar
	[2345] = { 54895 }, -- Ivus the Decayed
	[2329] = { 54896 }, -- Ivus the Forest Lord
	-- Nazjatar (BfA 8.2)
	[2362] = { 56057 }, -- Ulmath, the Soulbinder
	[2363] = { 56056 }, -- Wekemara
	-- Uldum and Vale corruptions (BfA 8.3)
	[2378] = { 58705 }, -- Grand Empress Shek'zara
	[2381] = { 55466 }, -- Vuk'laz the Earthbreaker
	-- Shadowlands
	[2430] = { 61813, 66614 }, -- Valinor, the Light of Eons
	[2431] = { 61816, 66617 }, -- Mortanis
	[2432] = { 61815, 66616 }, -- Oranomonos the Everbranching
	[2433] = { 61814, 66615 }, -- Nurgash Muckformed
	-- Korthia (Shadowlands 9.1)
	[2456] = { 64531, 66618 }, -- Mor'geth, Tormentor of the Damned
	-- Zereth Mortis (Shadowlands 9.2)
	[2468] = { 65143, 66619 }, -- Antros
	-- Dragonflight
	[2515] = { 69929 }, -- Strunraan, The Sky's Misery
	[2506] = { 69930 }, -- Basrikon, The Shale Wing
	[2517] = { 69927 }, -- Bazual, The Dreaded Flame
	[2518] = { 69928 }, -- Liskanoth, The Futurebane
	-- Zaralek Cavern (Dragonflight 10.1)
	[2531] = { 74892 }, -- The Zaqali Elders
	-- Emerald Dream
	[2562] = { 76367 }, -- Aurostor
}

-- set during init. key is the mapID/instanceID returned for an Encounter Journal
-- instance id, the value is an entry in the RaidFinder table.
local RaidFromInstance = {}

local RaidFinder = { -- raid id's are encounter journal ids, wing ids are LfgDungeonIDs

	--=========== DRAGONFLIGHT ===========--

	{ id = 1207, wings = { -- tier 31 - Amirdrassil, the Dream's Hope
		{ id = 2466, max = 3 }, -- Incarnate's Wake
		{ id = 2468, max = 2 }, -- Molten Incursion
		{ id = 2467, max = 2 }, -- The Viridian Weave
		{ id = 2469, max = 2 }, -- Fate of Amidrassil
	}},

	{ id = 1208, wings = { -- tier 30 - Aberrus, the Shadowed Crucible
		{ id = 2399, max = 3 }, -- Discarded Works
		{ id = 2400, max = 3 }, -- Fury of Giants
		{ id = 2401, max = 2 }, -- Neltharion's Shadow
		{ id = 2402, max = 1 }, -- Edge of the Void
	}},

	{ id = 1200, wings = { -- tier 29 - Vault of the Incarnates
		{ id = 2370, max = 3 }, -- The Primal Bulwark
		{ id = 2371, max = 3 }, -- Caverns of Infusion
		{ id = 2372, max = 2 }, -- Fury of the Storm
	}},


	--=========== SHADOWLANDS ============--

	{ id = 1195, wings = { -- tier 28 - Sepulcher of the First Ones
		{ id = 2292, max = 4 }, -- Ephemeral Plains
		{ id = 2291, max = 3 }, -- Cornerstone of Creation
		{ id = 2293, max = 3 }, -- Domination's Grasp
		{ id = 2294, max = 1 }, -- The Grand Design
	}},

	{ id = 1193, wings = { -- tier 27 - Sanctum of Domination
		{ id = 2221, max = 3 }, -- The Jailer's Vanguard
		{ id = 2222, max = 3 }, -- The Dark Bastille
		{ id = 2223, max = 3 }, -- Shackles of Fate
		{ id = 2224, max = 1 }, -- The Reckoning
	}},

	{ id = 1190, wings = { -- tier 26 - Castle Nathria
		{ id = 2090, max = 3 }, -- The Leeching Vaults
		{ id = 2091, max = 3 }, -- Reliquary of Opulence
		{ id = 2092, max = 3 }, -- Blood from Stone
		{ id = 2096, max = 1 }, -- An Audience with Arrogance
	}},


	--======== BATTLE FOR AZEROTH ========--

	{ id = 1180, wings = { -- tier 25 - Ny'alotha, the Waking City
		{ id = 2036, max = 3 }, -- Vision of Destiny
		{ id = 2037, max = 3 }, -- Halls of Devotion
		{ id = 2038, max = 3 }, -- Gift of Flesh
		{ id = 2039, max = 2 }, -- The Waking Dream
	}},

	{ id = 1179, wings = { -- tier 24 - The Eternal Palace
		{ id = 2009, max = 3 }, -- The Grand Reception
		{ id = 2010, max = 3 }, -- Depths of the Devoted
		{ id = 2011, max = 2 }, -- The Circle of Stars
	}},

	{ id = 1177, wings = { -- tier 23 - Crucible of Storms
		{ id = 1951, max = 2 }, -- Crucible of Storms
	}},

	{ id = 1176, wings = { -- tier 23 - Battle of Dazar'alor
		setmetatable( { max = 3 }, { __index = function(t,k) if k == "id" then return ("Alliance" == UnitFactionGroup("player")) and 1945 or 1948 end end } ), -- Siege of Dazar'alor OR Defense of Dazar'alor
		setmetatable( { max = 3 }, { __index = function(t,k) if k == "id" then return ("Alliance" == UnitFactionGroup("player")) and 1946 or 1949 end end } ), -- Empire's Fall OR Death's Bargain
		setmetatable( { max = 3 }, { __index = function(t,k) if k == "id" then return ("Alliance" == UnitFactionGroup("player")) and 1947 or 1950 end end } ), -- Might of the Alliance OR Victory or Death
	}},

	{ id = 1031, wings = { -- tier 22 - Uldir
		{ id = 1731, max = 3 }, -- Halls of Containment
		{ id = 1732, max = 3 }, -- Crimson Descent
		{ id = 1733, max = 2 }, -- Heart of Corruption
	}},


	--============== LEGION ==============--
	
	{ id = 946, wings = { -- tier 21 - Antorus, the Burning Throne
		{ id = 1610, max = 3 }, -- Light's Breach
		{ id = 1611, max = 3 }, -- Forbidden Descent
		{ id = 1612, max = 3 }, -- Hope's End
		{ id = 1613, max = 2 }, -- Seat of the Pantheon
	}},
	{ id = 875, wings = { -- tier 20 - Tomb of Sargeras
		{ id = 1494, max = 3 }, -- The Gates of Hell
		{ id = 1495, max = 3 }, -- Wailing Halls
		{ id = 1496, max = 2 }, -- Chamber of the Avatar
		{ id = 1497, max = 1 }, -- Deceiver's Fall
	}},

	{ id = 861, wings = { -- non-tier - Trial of Valor
		{ id = 1411, max = 3 }, -- Trial of Valor
	}},

	{ id = 786, wings = { -- tier 19 - The Nighthold
		{ id = 1290, max = 3 }, -- Arcing Aqueducts
		{ id = 1291, max = 3 }, -- Royal Athenaem
		{ id = 1292, max = 3 }, -- Nightspire
		{ id = 1293, max = 1 }, -- Betrayer's Rise
	}},
	{ id = 768, wings = { -- tier 19 - The Emerald Nightmare
		{ id = 1287, max = 3 }, -- Darkbough
		{ id = 1288, max = 3 }, -- Tormented Guardians
		{ id = 1289, max = 1 }, -- Rift of Aln
	}},

	
	--============== WARLORDS OF DRAENOR ==============--

	{ id = 669, wings = { -- tier 18 - Hellfire Citadel
		{ id = 982,  max = 3 }, -- Hellbreach
		{ id = 983,  max = 3 }, -- Halls of Blood
		{ id = 984,  max = 3 }, -- Bastion of Shadows
		{ id = 985,  max = 3 }, -- Destructor's Rise
		{ id = 986,  max = 1 }, -- The Black Gate
	}},

	{ id = 457, wings = { -- tier 17 - Black Rock Foundry
		{ id = 847,  max = 3 }, -- Slagworks
		{ id = 846,  max = 3 }, -- The Black Forge
		{ id = 848,  max = 3 }, -- Iron Assembly
		{ id = 823,  max = 1 }, -- Blackhand's Crucible
	}},
	{ id = 477, wings = { -- tier 17 - Highmaul
		{ id = 849,  max = 3 }, -- Walled City
		{ id = 850,  max = 3 }, -- Arcane Sanctum
		{ id = 851,  max = 1 }, -- Imperator's Rise
	}},


	-- ============== MISTS OF PANDARIA ==============--

	{ id = 369, wings = { -- tier 16 - Siege of Orgrimmar
		{ id = 716, max = 4 }, -- Vale of Eternal Sorrows
		{ id = 717, max = 4 }, -- Gates of Retribution
		{ id = 724, max = 3 }, -- The Underhold
		{ id = 725, max = 3 }, -- Downfall
	}},

	{ id = 362, wings = { -- tier 15 - Throne of Thunder
		{ id = 610, max = 3 }, -- Last Stand of the Zandalari
		{ id = 611, max = 3 }, -- Forgotten Depths
		{ id = 612, max = 3 }, -- Halls of Flesh-Shaping
		{ id = 613, max = 3 }, -- Pinnacle of Storms
	}},

	{ id = 320, wings = { -- tier 14 - Terrace of Endless Spring
		{ id = 526, max = 4 }, -- Terrace of Endless Spring
	}},
	{ id = 330, wings = { -- tier 14 - Heart of Fear
		{ id = 529, max = 3 }, -- The Dread Approach
		{ id = 530, max = 3 }, -- Nightmare of Shek'zeer
	}},

	{ id = 317, wings = { -- non-tier - Mogu'shan Vaults
		{ id = 527, max = 3 }, -- Guardians of Mogu'shan
		{ id = 528, max = 3 }, -- The Vault of Mysteries
	}},


	--============== CATACLYSM ==============--

	{ id = 187, wings = { -- tier 13 - Dragonsoul
		{ id = 416, max = 4 }, -- The Siege of Wyrmrest Temple
		{ id = 417, max = 4 }, -- Fall of Deathwing
	}},
}


local LegacyRaid = {
	--============== CATACLYSM ==============--

	{ tier = 4, raids = {
		{ id = 78 }, -- tier 12 - Firelands

		{ id = 74 }, -- tier 11 - Throne of Four Winds
		{ id = 72 }, -- tier 11 - Bastion of Twilight
		{ id = 73 }, -- tier 11 - Blackwing Descent

		{ id = 75 }, -- tier 11-13 - Baradin Hold
	}},


	--============== WRATH OF THE LICH KING ==============--

	{ tier = 3, raids = {
		{ id = 761 }, -- tier 10 - The Ruby Sanctum
		{ id = 758 }, -- tier 10 - Icecrown Citadel

		{ id = 760 }, -- tier 9 - Onyxia's Lair
		{ id = 757 }, -- tier 9 - Trial of the Crusader

		{ id = 759 }, -- tier 8 - Ulduar

		{ id = 756 }, -- tier 7 - The Eye of Eternity
		{ id = 755 }, -- tier 7 - The Obsidian Sanctum
		{ id = 754 }, -- tier 7 - Naxxramas

		{ id = 753 }, -- tier 7-10 - Vault of Archavon
	}},


	--============== BURNING CRUSADE ==============--

	{ tier = 2, raids = {
		-- Raid lockout is saved as The Sunwell but appears in Encounter Journal as Sunwell Plateau.
		-- There is a map id with the name The Sunwell, so we can try using that instead.
		-- We'll add both versions, though, because other languages might be correct for all we know.
		{map = 973 }, -- tier 6.5 - The Sunwell
		{ id = 752 }, -- tier 6.5 - Sunwell Plateau

		{ id = 751 }, -- tier 6 - Black Temple
		{ id = 750 }, -- tier 6 - The Battle Mount Hyjal

		{map = 334 }, -- tier 5 - Tempest Keep
		{ id = 749 }, -- tier 5 - The Eye
		{ id = 748, name = L["LOCKOUT_Serpentshrine_Cavern"] }, -- tier 5 - Serpentshrine Cavern

		{ id = 747 }, -- tier 4 - Magtheridon's Lair
		{ id = 746 }, -- tier 4 - Gruul's Lair
		{ id = 745 }, -- tier 4 - Karazhan
	}},


	--============== CLASSIC ==============--

	{ tier = 1, raids = {
		{ id = 744, name = L["AQT_LOCKOUT_NAME"] }, -- tier 2.5 - Temple of Ahn'Qiraj
		{ id = 743 }, -- tier 2.5 - Ruins of Ahn'Qiraj

		{ id = 742 }, -- tier 2 - Blackwing Lair

		{ id = 741, name = L["LOCKOUT_MOLTEN_CORE"] }, -- tier 1 - Molten Core
	}},

}



--=========================================================================--
--
-- these tables are filled in during initialization using data from above
--
--=========================================================================--

local FactionRenown = {}

local Dungeons = {}

local RepeatableQuests = {} -- used to build TrackedQuests and TrackedCurrencies during InitQuestsTable
local TrackedQuests = {} -- structure is { questId = true }

local Currencies = {} -- used to build TrackedCurrencies during InitCurrenciesTable, referenced during Tooltip layout
local TrackedCurrencies = {} -- structure is { currencyId = { tracker, Get, item, name, total, weekly, (optional) totalMax, (optional) weeklyMax | |
                      -- OR { currencyId = { seasonal = true, tracker, Get, item, name, total, earned, seasonMax } }
                      -- OR { currencyId = { trigger  = true, tracker, Get, item, name, total } }

local OrderedColumns = {}

--[===[@debug@
Ailo.FactionRenown = FactionRenown
Ailo.Dungeons = Dungeons
Ailo.RepeatableQuests = RepeatableQuests
Ailo.TrackedQuests = TrackedQuests
Ailo.Currencies = Currencies
Ailo.TrackedCurrencies = TrackedCurrencies
Ailo.OrderedColumns = OrderedColumns
--@end-debug@]===]


--=========================================================================--
--=========================================================================--
--
-- Addon Initialization
--
--=========================================================================--
--=========================================================================--

function Ailo:OnInitialize() -- ADDON_LOADED, only saved variables are guaranteed to be present
	defaults.global.weeklyResetDay = GetWeeklyResetDay()
	self.db = LibStub("AceDB-3.0"):New("AiloDB", defaults, true)
	self.db.RegisterCallback(self, "OnProfileChanged", "MigrateSavedData")
	self:MigrateSavedData()

	LibStub("AceConfig-3.0"):RegisterOptionsTable("Ailo", self.GenerateOptions)
	self.optionsFrame = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("Ailo")

	local AiloLDB = ldb:NewDataObject("Ailo", {
		type = "data source",
		text = "Ailo",
		icon = "Interface\\Icons\\Achievement_Dungeon_UlduarRaid_Archway_01.png",
		OnClick = function(clickedframe, button)
			if button == "RightButton" then 
				_G.InterfaceOptionsFrame_OpenToCategory(Ailo.optionsFrame) 
			else 
				if _G.IsShiftKeyDown() then
					Ailo:ManualPlayerUpdate() 
				else
					_G.ToggleFriendsFrame(4)
				end
			end
		end,
		OnEnter = function(tt)
			local tooltip = LibQTip:Acquire("AiloTooltip", 1, "LEFT") 
			Ailo.tooltip = tooltip
			Ailo:PrepareTooltip(tooltip) 
			tooltip:SmartAnchorTo(tt)
			tooltip:Show()
		end,
		OnLeave = function(tt)
			LibQTip:Release(Ailo.tooltip)
			Ailo.tooltip = nil
		end,
	})

	LDBIcon:Register("Ailo", AiloLDB, self.db.profile.minimapIcon)
end

function Ailo:OnEnable() -- PLAYER_LOGIN, game data is initialized
	charLevel = UnitLevel("player")
	currentRealmChar = strformat("%s.%s", GetRealmName(), UnitName("player"))
	strfDBKey = strformat("%s.%%s", currentRealmChar)  -- ie Stormrage.Bob.%s

	self:SetupClasscoloredFonts()
	if CUSTOM_CLASS_COLORS then
		CUSTOM_CLASS_COLORS:RegisterCallback("SetupClasscoloredFonts", self)
	end

	local fontname, fontheight, fontflags
	local tooltip = LibQTip:Acquire("AiloTooltip", 1, "LEFT")

	CellFonts.iconTable = CreateFont("AiloFontIconTable")
	fontname, fontheight, fontflags = tooltip:GetFont():GetFont()
	CellFonts.iconTable:SetFont(fontname, fontheight * 1.25, fontflags)
	CellFonts.iconTable:SetTextColor(1, 1, 1)

	CellFonts.iconHeader = CreateFont("AiloFontIconHeader")
	local fontname, fontheight, fontflags = tooltip:GetHeaderFont():GetFont()
	CellFonts.iconHeader:SetFont(fontname, fontheight * 1.25, fontflags)
	CellFonts.iconHeader:SetTextColor(1, 1, 1)

	LibQTip:Release(tooltip)

	updateRetryTimer = self:ScheduleRepeatingTimer("DelayedInit", .2)
end

function Ailo:LevelCheck(info, newMinLevel)
	self.db.profile.minLevel = newMinLevel
end

function Ailo:DelayedInit()
	-- When GetNextDailyReset is called early, like during OnEnable, it can 
	-- return hugeNumber. If a hugeNumber is saved as an expire time acedb
	-- will convert it to a nil. This is most often a problem for currency
	-- trackers which could only check once, during the first UpdatePlayer
	-- called during OnEnable. Solution is to use an Ace Timer callback to
	-- introduce a small delay between OnEnable and PlayerUpdate to give 
	-- the game time get the daily quest reset time sorted out.
	-- If we hit RetryLimit, stop delaying and just let it init.  --REVIEW - how problematic is that?

	if updateRetryCount < updateRetryLimit then
		updateRetryCount = updateRetryCount + 1
		if GetNextDailyReset() == hugeNumber then
			return
		end
	end

	if updateRetryTimer then
		self:CancelTimer(updateRetryTimer, true)
		updateRetryCount = updateRetryLimit
		updateRetryTimer = nil
	end

	-- delay means we should be able to get all the strings and values needed for initialization
	self:InitTrackersTable()
	self:InitQuestsTable()
	self:InitCurrenciesTable()  -- InitQuestsTable must happen before this
	self:InitDungeonsTable()
	self:InitRaidFinderTable()
	self:InitLegacyRaidTable()

	-- we are now ready to respond to update events
	self:RegisterEvent("LFG_LIST_SEARCH_RESULT_UPDATED",        "ColorLFGSearchResults")
	self:RegisterEvent("ZONE_CHANGED_NEW_AREA",                 "CheckCurrentMap")
	self:RegisterEvent("COVENANT_CALLINGS_UPDATED",             "UpdateCovenantCallings")  -- TriggerCovenantCallings will prompt this event
	self:RegisterEvent("UPDATE_INSTANCE_INFO",                  "UpdateLockouts")          -- TriggerLockoutsUpdate will prompt this event
	self:RegisterEvent("LFG_UPDATE_RANDOM_INFO",                "UpdateLFGResets")         -- TriggerLFGResetsUpdate will prompt this event
	self:RegisterEvent("CHALLENGE_MODE_MAPS_UPDATE",            "UpdateMythicPlus")        -- UpdateMythicPlus will trigger this the first time it is called


	self:RegisterEvent("PLAYER_REGEN_ENABLED",                  "CentralDispatch")
	self:RegisterEvent("PLAYER_LEVEL_UP",                       "CentralDispatch")  -- UpdateRoster (includes UpdateRenown)
	self:RegisterEvent("COVENANT_SANCTUM_RENOWN_LEVEL_CHANGED", "CentralDispatch")  -- UpdateRoster (includes UpdateRenown)
	self:RegisterEvent("MAJOR_FACTION_UNLOCKED",                "CentralDispatch")  -- UpdateRenown
	self:RegisterEvent("MAJOR_FACTION_RENOWN_LEVEL_CHANGED",    "CentralDispatch")  -- UpdateRenown
	self:RegisterEvent("CURRENCY_DISPLAY_UPDATE",               "CentralDispatch")  -- UpdateCurrencies + (sometimes) UpdateQuests
	self:RegisterEvent("CHAT_MSG_SYSTEM",                       "CentralDispatch")  -- TriggerLockoutsUpdate
	self:RegisterEvent("SHOW_LOOT_TOAST",                       "CentralDispatch")  -- TriggerLockoutsUpdate + UpdateQuests + UpdateRenown + TriggerCovenantCallings
	self:RegisterEvent("LFG_COMPLETION_REWARD",                 "CentralDispatch")  -- TriggerLFGResetsUpdate
	self:RegisterEvent("WEEKLY_REWARDS_UPDATE",                 "CentralDispatch")  -- UpdateMythicPlus + UpdateMythicPlusKey
	self:RegisterEvent("QUEST_LOG_UPDATE",                      "CentralDispatch")  -- UpdateEmissaryBounties
	self:RegisterEvent("UNIT_QUEST_LOG_CHANGED",                "CentralDispatch")  -- UpdateQuests + UpdateCurrencies + TriggerCovenantCallings
	self:RegisterEvent("QUEST_TURNED_IN",                       "CentralDispatch")  -- UpdateQuests + UpdateCurrencies + TriggerCovenantCallings
	self:RegisterEvent("CRAFTINGORDERS_FULFILL_ORDER_RESPONSE", "CentralDispatch")  -- UpdateQuests + UpdateCurrencies + TriggerCovenantCallings
	self:RegisterEvent("SCENARIO_COMPLETED",                    "CentralDispatch")  -- UpdateQuests + UpdateCurrencies + TriggerCovenantCallings


	-- trigger collection of lockout, reset and quest information
	self:PurgeExpiredData()
	self:UpdatePlayer()
end

function Ailo:OnDisable()
	self:CancelAllTimers()

	self:UnregisterEvent("CHAT_MSG_SYSTEM")
	self:UnregisterEvent("CURRENCY_DISPLAY_UPDATE")
	self:UnregisterEvent("LFG_COMPLETION_REWARD")
	self:UnregisterEvent("LFG_UPDATE_RANDOM_INFO")
	self:UnregisterEvent("PLAYER_LEVEL_UP")
	self:UnregisterEvent("SHOW_LOOT_TOAST")
	self:UnregisterEvent("UNIT_QUEST_LOG_CHANGED")
	self:UnregisterEvent("QUEST_TURNED_IN")
	self:UnregisterEvent("UPDATE_INSTANCE_INFO")
	self:UnregisterEvent("ZONE_CHANGED_NEW_AREA")
	self:UnregisterEvent("CHALLENGE_MODE_COMPLETED")
	self:UnregisterEvent("CHALLENGE_MODE_MAPS_UPDATE")
	self:UnregisterEvent("WEEKLY_REWARDS_UPDATE")
	self:UnregisterEvent("QUEST_LOG_UPDATE")
	self:UnregisterEvent("COVENANT_CALLINGS_UPDATED")
	self:UnregisterEvent("COVENANT_SANCTUM_RENOWN_LEVEL_CHANGED")
	self:UnregisterEvent("MAJOR_FACTION_UNLOCKED")
	self:UnregisterEvent("MAJOR_FACTION_RENOWN_LEVEL_CHANGED")
end

-- if we change the structure or contents of the saved settings, we can 
-- do a fixup of the data here the first time we run after the upgrade
function Ailo:MigrateSavedData()
	local  globaldb = self.db.global
	local profiledb = self.db.profile

	-- all db sections are current we can exit immediately
	if profiledb.version == DB_VERSION and
	    globaldb.version == DB_VERSION then
		return
	end

	-- migration code makes use of calls to WipeDB,
	-- this is version of WipeDB that existed before db version 16
	local function Pre16WipeDB()
		if type(globaldb.chars) == "table" then wipe(globaldb.chars) end
		if type(globaldb.raids) == "table" then wipe(globaldb.raids) end
		if type(globaldb.lfrs)  == "table" then wipe(globaldb.lfrs)  end
		if type(globaldb.world) == "table" then wipe(globaldb.world) end
		if type(globaldb.charClass) == "table" then wipe(globaldb.charClass) end
		globaldb.nextPurge = GetNextDailyReset()
	end

	-- move/rename a db value
	local function Move(oldDB, oldKey, newDB, newKey)
		local value = oldDB[oldKey]
		if value ~= nil then
			newDB[newKey] = value
			oldDB[oldKey] = nil
		end
	end


	-- prior to DB_VERSION == 16, stamping was incorrect in that only the global was stamped and
	-- that value was used to determine whether any profile migration should occur. that meant
	-- only the first profile loaded after upgrade got the migration. if they only used one profile,
	-- then there won't be any problems, but we'll repeat key nukes for "new" profiles anyway
	if not profiledb.version then
		-- options removed from the addon before profile versioning was correctly set
		profiledb.showWeeklyRaid = nil
		profiledb.showDailyHeroic = nil
		profiledb.showShaOfAnger = nil
		profiledb.showGalleon = nil
		profiledb.showNalak = nil
		profiledb.showOondasta = nil
		profiledb.showWeeklyValor = nil
		profiledb.showFlexRaid = nil
		profiledb.hc = nil
		profiledb.nhc = nil

		profiledb.version = DB_VERSION
	end

	-- if there is no version stamp, only safe thing to do is wipe it
	if not globaldb.version then
		Pre16WipeDB()
		Ailo:WipeDB()
		globaldb.version = DB_VERSION
	end


	-- if version is less current, settings may need to be migrated/tweaked/etc
	if globaldb.version < 9 then
		-- change to values stored for daily heroic
		self:Output(L["DB_VERSION_UPGRADE_PURGE"])
		Pre16WipeDB()

		-- if they weren't showing the old VP column, don't show the new
		profiledb.showWeeklyValor = profiledb.showDailyHeroic
	end

	if globaldb.version < 10 then
		-- database does not need to be wiped for these changes

		-- if they were showing the daily PVP column, they'll likely want weekly CP tracking
		profiledb.showWeeklyConquest = profiledb.showDailyPVP
		-- remove dalaran weekly raid quest tracking
		profiledb.showWeeklyRaid = nil
	end

	if globaldb.version < 11 then
		-- change to values stored for naming convention normalization
		self:Output(L["DB_VERSION_UPGRADE_PURGE"])
		Pre16WipeDB()

		-- name change to distinguish from daily heroic scenario
		Move(profiledb, "showDailyHeroic", profiledb, "showDailyHeroicDungeon")

		-- show daily heroic scenario if already showing daily heroic dungeon or daily scenario
		profiledb.showDailyHeroicScenario = profiledb.showDailyHeroicDungeon or profiledb.showDailyScenario
	end

	if globaldb.version < 12 then
		-- only one toggle for world bosses now with blizz api change
		-- database does not need to be wiped for these changes
		profiledb.showShaOfAnger = nil
		profiledb.showGalleon = nil
		profiledb.showNalak = nil
		profiledb.showOondasta = nil
	end

	if globaldb.version < 13 then
		self:Output(L["DB_VERSION_UPGRADE_PURGE"])
		Pre16WipeDB()
	end

	if globaldb.version < 14 then
		-- valor points no longer exist
		profiledb.showWeeklyValor = nil
		self:Output(L["DB_VERSION_UPGRADE_PURGE"])
		Pre16WipeDB()
	end

	if globaldb.version < 15 then
		-- flex raids no longer exist
		profiledb.showFlexRaid = nil
		self:Output(L["DB_VERSION_UPGRADE_PURGE"])
		Pre16WipeDB()
	end

	if globaldb.version < 16 then
			self:Output(L["DB_VERSION_UPGRADE_PURGE"])
			globaldb.chars = nil
			globaldb.charClass = nil
			globaldb.currencies = nil
			globaldb.raids = nil
			globaldb.lfrs = nil
			globaldb.world = nil
	end

	--=========================================================================--
	-- from here on, the dbversion check and setting location should match
	--=========================================================================--

	if profiledb.version < 17 then
		profiledb.difficultyAbbr[strTimewalker] = AbbreviateString(strTimewalker)
	end

	if profiledb.version < 18 then
		--   added new values to defaults (no migration needed, just documenting the change)
		--     showHellbaneRares = true,
	end

	if profiledb.version < 19 then
		-- we were using both Conquest and WeeklyConquest as names for same thing. standardize
		Move(profiledb, "showWeeklyConquest", profiledb, "showConquest")

		-- some spanner and column names were being inserted into the instance abbr's
		-- delete them, but retain any user customization
		local function MoveInstanceAbbr(name, dest, key)
			local value = profiledb.instanceAbbr[name]
			if value ~= nil then
				if value ~= AbbreviateString(name) then
					dest[key] = value
				end
				profiledb.instanceAbbr[name] = nil
			end
		end
		MoveInstanceAbbr(strMythic, profiledb.difficultyAbbr, strMythic)
		MoveInstanceAbbr("Bounties", profiledb.columnName, "bounties")
		MoveInstanceAbbr("World Bosses", profiledb.columnName, "worldBosses")
	end

	if globaldb.version < 20 then
		-- N'Zoth Assaults moved from special assaults section into weekly and biweekly storage
		if type(globaldb.assaults) == "table" then
			for key, data in pairs(globaldb.assaults) do
				data.reset = nil
				if strfind(key, "Empire", 1, true) then
					globaldb.weeklies[strgsub(key, "NZoth", "NZothAssaults")] = data
				else
					globaldb.biweeklies[strgsub(key, "NZoth", "NZothAssaults")] = data
				end
			end
			globaldb.assaults = nil
		end
	end

	if profiledb.version < 20 then
		-- Dalaran tracker is now WeeklyEvent
		Move(profiledb, "showDalaran", profiledb, "showWeeklyEvent")
	end

	if globaldb.version < 21 then
		-- changes to how we store bounty data
		local translate = {}
		for k,v in pairs(bounty_list_bfa) do
			translate[tostring(v.icon)] = k
		end

		local db = globaldb.bounties
		for k,v in pairs(db) do
			if k ~= "timers" and k ~= "resets" then
				local _, _, realmChar, icon = breakKey(k)
				local newkey = translate[icon]
				if newkey then
					db.resets[strformat("%s.%s", realmChar, newkey)] = v
				end
				db[k] = nil
			end
		end
	end

	if profiledb.version < 22 then
		-- move all the showTracker flags to show.Tracker
		Move(profiledb, "show5Man",                profiledb.show, "5Man")
		Move(profiledb, "showDailyHeroicScenario", profiledb.show, "DailyHeroicScenario")
		Move(profiledb, "showDailyScenario",       profiledb.show, "DailyScenario")
		Move(profiledb, "showRaidFinder",          profiledb.show, "RaidFinder")
		Move(profiledb, "showWorldBosses",         profiledb.show, "WorldBosses")
		Move(profiledb, "showTBVictory",           profiledb.show, "TBVictory")
		Move(profiledb, "showWGVictory",           profiledb.show, "WGVictory")
		Move(profiledb, "showConquest",            profiledb.show, "Conquest")
		Move(profiledb, "showSupplyNazjatar",      profiledb.show, "SupplyNazjatar")
		Move(profiledb, "showWeeklyNazjatar",      profiledb.show, "WeeklyNazjatar")
		Move(profiledb, "showDailyPVP",            profiledb.show, "DailyPVP")
		Move(profiledb, "showSeasonal",            profiledb.show, "Seasonal")
		Move(profiledb, "showTillers",             profiledb.show, "Tillers")
		Move(profiledb, "showLootCoins",           profiledb.show, "LootCoins")
		Move(profiledb, "showGarrisonInvasion",    profiledb.show, "GarrisonInvasion")
		Move(profiledb, "showScrapMeltdown",       profiledb.show, "ScrapMeltdown")
		Move(profiledb, "showHellbaneRares",       profiledb.show, "HellbaneRares")
		Move(profiledb, "showMythicPlus",          profiledb.show, "MythicPlus")
		Move(profiledb, "showIslands",             profiledb.show, "Islands")
		Move(profiledb, "showNZothAssaults",       profiledb.show, "NZothAssaults")
		Move(profiledb, "showBounties",            profiledb.show, "Bounties")
		Move(profiledb, "showWeeklyEvent",         profiledb.show, "WeeklyEvent")

		-- columnName really should have been keyed by tracker name; case was wrong
		Move(profiledb.columnName, "dailies",          profiledb.columnName, "Dailies")
		Move(profiledb.columnName, "weeklies",         profiledb.columnName, "Weeklies")
		Move(profiledb.columnName, "bounties",         profiledb.columnName, "Bounties")
		Move(profiledb.columnName, "assaults",         profiledb.columnName, "NZothAssaults")
		Move(profiledb.columnName, "worldBosses",      profiledb.columnName, "WorldBosses")
		Move(profiledb.columnName, "total",            profiledb.columnName, "Total")
		Move(profiledb.columnName, "weekly",           profiledb.columnName, "Weekly")
		Move(profiledb.columnName, "conquest",         profiledb.columnName, "Conquest")
		Move(profiledb.columnName, "mythicPlus",       profiledb.columnName, "MythicPlus")
		Move(profiledb.columnName, "garrisonInvasion", profiledb.columnName, "GarrisonInvasion")
	end

	if profiledb.version < 23 then
		-- after the level squish, a custom value for minLevel may be above the new max level
		if profiledb.minLevel > 50 then
			-- lets try to keep showing the non-max characters they had been seeing
			-- we'll keep half of their previous difference
			local diff = (120 - profiledb.minLevel) / 2
			profiledb.minLevel = max(1, 50-diff)
		end
	end

	if globaldb.version < 24 then
		-- we understand seasonal currencies now; blow away old/bad data in total/weekly
		local db = globaldb.currencies.total
		for k,v in pairs(db) do
			local _, _, _, currency = breakKey(k)
			if currency == "Valor" or currency == "Conquest" then
				db[k] = nil
			end
		end
		local db = globaldb.currencies.weekly
		for k,v in pairs(db) do
			local _, _, _, currency = breakKey(k)
			if currency == "Valor" or currency == "Conquest" then
				db[k] = nil
			end
		end
	end

	if globaldb.version < 25 then
		-- seasonal currency save data changing
		-- instead of saving tracker name we will save currency id
		local map = { Valor = 1191, Conquest = 1602, TowerKnowledge = 1904 }
		local db = globaldb.currencies.season
		for k,v in pairs(db) do
			v.id = v.key and map[v.key]
			v.key = nil
		end
	end

	if globaldb.version < 26 then
		-- restructure covenant renown in prep for adding dragon flight renown tracking
		for charRealm, data in pairs(globaldb.roster) do
			local level = data.renown

			data.renown = {}
			if level then
				data.renown.Covenant = (level == 80) and "done" or level
			end
		end
	end

	if profiledb.version < 26 then
		-- restructure covenant renown in prep for adding dragon flight renown tracking
		if rawget(profiledb.show, "Renown") then
			profiledb.show.Covenant = true
		end
		profiledb.show.Renown = nil

		profiledb.columnName.Covenant = nil
		profiledb.columnName.Renown = L["Renown"]
	end

	if globaldb.version < 27 then
		local weeklies = globaldb.weeklies
		for k,v in pairs(weeklies) do
			local _, _, realmChar, name = breakKey(k)
			if name == "MythicPlus" then
				v.level = v.count
				v.count = nil
				v.max = nil

				weeklies[realmChar .. "." .. "MythicPlusReward1"] = v
				weeklies[k] = nil
			end
		end
	end

	if profiledb.version < 27 then
		-- mythic plus rework
		if false == rawget(profiledb.show, "MythicPlus") then
			profiledb.show.MythicPlusReward1 = false
			profiledb.show.MythicPlusReward2 = false
			profiledb.show.MythicPlusReward3 = false
			profiledb.show.MythicPlus = nil
		end
	end

	if globaldb.version < 28 then
		-- Dragonflight S1 catalyst charges still exist as a currency, but aren't spendable
		-- we're going to change the tracker so remove old data
		local total = globaldb.currencies.total
		for k,v in pairs(total) do
			local _, _, _, name = breakKey(k)
			if name == "CatalystCharges" then
				total[k] = nil
			end
		end

		-- Archives quest did not reset completion upon end of season 1
		-- we will no longer look for them, but lets not leave junk data behind
		-- do a one time removal
		local weeklies = globaldb.weeklies
		for k,v in pairs(weeklies) do
			local _, _, _, name = breakKey(k)
			if name == "ArchivesDungeon" then
				weeklies[k] = nil
			end
		end
	end

	if globaldb.version < 29 then
		-- Flightstones are a simply capped currency, not seasonally evolving
		local season = globaldb.currencies.season
		for k,v in pairs(season) do
			local _, _, _, name = breakKey(k)
			if name == "Flightstones" then
				season[k] = nil
			end
		end
	end

	-- WeeklyEvent broken up into seperate trackers to better handle multiple events in one week
	-- also, seperate icons mean you can identify which weekly it is
	if globaldb.version < 30 then
		local weeklies = globaldb.weeklies
		for k,v in pairs(weeklies) do
			local _, _, _, name = breakKey(k)
			if name == "WeeklyEvent" then
				weeklies[k] = nil
			end
		end
	end
	if profiledb.version < 30 then
		if false == rawget(profiledb.show, "WeeklyEvent") then
			profiledb.show.WeeklyTheWorldAwaits   = false
			profiledb.show.WeeklyAPathThroughTime = false
			profiledb.show.WeeklyEmmissaryOfWar   = false
			profiledb.show.WeeklyTheVeryBest      = false
			profiledb.show.WeeklyACallToBattle    = false
			profiledb.show.WeeklyTheArenaCalls    = false
			profiledb.show.WeeklyEvent = nil
		end
	end

	-- change to how we display profession knowledge trackers
	if globaldb.version < 32 then
		-- this is unfortuneate, but we would display wrong
		-- stuff for characters you haven't logged into yet
		self:WipeDB()
	end

	-- reset seasonal currencies for 10.2
	if globaldb.version < 33 then
		globaldb.currencies.season = {}

		total = globaldb.currencies.total
		for k,v in pairs(total) do
			local _, _, _, name = breakKey(k)
			if name == "Flightstones" then
				total[k] = nil
			end
		end
	end

	-- Enum.WeeklyRewardChestThresholdType.MythicPlus was replaced by Enum.WeeklyRewardChestThresholdType.Activities
	-- and that produced some bad data we need to clean up
	if globaldb.version < 34 then
		local weeklies = globaldb.weeklies
		for k,v in pairs(weeklies) do
			local _, _, i = strfind(k, "MythicPlusReward(%d+)")
			if i and not (i == "1" or i == "2" or i == "3") then
				weeklies[k] = nil
			end
		end
	end

	-- catalyst charge id change for 10.2
	if globaldb.version < 35 then
		total = globaldb.currencies.total
		for k,v in pairs(total) do
			local _, _, _, name = breakKey(k)
			if name == "CatalystCharges" then
				total[k] = nil
			end
		end
	end

	profiledb.version = DB_VERSION
	 globaldb.version = DB_VERSION
end

function Ailo:InitTrackersTable()
	local columnName = self.db.profile.columnName

	local function CopyCriteria(tracker)
		local key = tracker.key
		Currencies[key] = tracker.currency
		RepeatableQuests[key] = tracker.quest or tracker.skillQuest
		Dungeons[key] = tracker.dungeon
		FactionRenown[key] = tracker.renown and tracker.renown.id
	end

	local function DoChildren(children, columns)
		for _, tracker in ipairs(children) do
			if tracker.children then
				-- anything between the spanners and the leaves of the tree are option groupings
				-- we do not create columns for those, only for their children
				DoChildren(tracker.children, columns)
			else
				-- we can assume all columns are keyed, if not the definition is bad
				CopyCriteria(tracker)
				local column = { key = tracker.key, showEmpty = tracker.showEmpty, format = tracker.format and CellFormat[tracker.format] }
				if tracker.abbr then
					column.abbr = tracker.abbr
					column.font = tracker.font and CellFonts[tracker.font]
				elseif tracker.columnName then
					column.abbr = columnName[tracker.columnName]
					column.font = tracker.font and CellFonts[tracker.font]
				elseif tracker.encounter then
					local name = EJ_GetInstanceInfo(tracker.encounter)
					local abbr = name and self:GetInstanceAbbr(name) or "error"
					if tracker.icon then
						column.abbr = strformat("|T%i:0|t%s", tracker.icon, abbr)
					else
						column.abbr = abbr
					end
				elseif tracker.icon then
					column.abbr = strformat("|T%s:0|t", tracker.icon)
					column.font = tracker.font and CellFonts[tracker.font] or CellFonts.iconHeader
				elseif tracker.torgWing then
					local name = C_ScenarioInfo.GetJailersTowerTypeString(tracker.torgWing)
					local abbr = name and self:GetInstanceAbbr(name) or "error"
					column.abbr = strformat("|T3743738:0|t%s", abbr) -- soul ash icon before the abbr
				end

				columns[#columns + 1] = column
			end
		end
	end

	for _, group in ipairs(Trackers) do
		if group.key then
			-- this check is only for LootCoins or any other currency split into multiple columns
			CopyCriteria(group)
		end

		local spanner = { fullname = group.fullname, columns = {} }
		if group.abbr then
			spanner.abbr = group.abbr
			spanner.font = group.font and CellFonts[group.font]
		elseif group.columnName then
			spanner.abbr = columnName[group.columnName]
			spanner.font = group.font and CellFonts[group.font]
		elseif group.icon then
			spanner.abbr = strformat("|T%s:0|t", group.icon)
			spanner.font = group.font and CellFonts[group.font] or CellFonts.iconHeader
		end

		DoChildren(group.children, spanner.columns)
		OrderedColumns[#OrderedColumns + 1] = spanner
	end
end

function Ailo:InitDungeonsTable()
	-- Let the game tell us the dungeon ID to use for this character
	Dungeons.DailyHeroicDungeon.id = GetRandomDungeonBestChoice() or 1671

	for name, dungeon in pairs(Dungeons) do
		local minLevel, maxLevel = select(LFG_RETURN_VALUES.minLevel, GetLFGDungeonInfo(dungeon.id))
		dungeon.minLevel = minLevel
		dungeon.maxLevel = maxLevel

		dungeon.first = dungeon.first or 0
		dungeon.last  = dungeon.last  or hugeNumber
	end
end

function Ailo:InitRaidFinderTable()
	for _, raid in ipairs(RaidFinder) do
		local name, _, _, _, _, _, _, _, _, dungeonId = EJ_GetInstanceInfo(raid.id)
		raid.name = strgsub(name or "error", "%.", "")
		for _, wing in ipairs(raid.wings) do
			wing.name = strgsub(GetLFGDungeonInfo(wing.id), "%.", "")
			wing.level = select(LFG_RETURN_VALUES.minLevel, GetLFGDungeonInfo(wing.id))
		end

		RaidFromInstance[dungeonId] = raid
	end
end

function Ailo:InitLegacyRaidTable()
	for _, expac in ipairs(LegacyRaid) do
		expac.name = EJ_GetTierInfo(expac.tier)
		for _, raid in ipairs(expac.raids) do
			if not raid.name or raid.name == "EJ" then
				if raid.map then
					local map = cmp.GetMapInfo(raid.map)
					raid.name = strgsub((map and map.name) or "error", "%.", "")
				else
					raid.name = strgsub(EJ_GetInstanceInfo(raid.id) or "error", "%.", "")
				end
			end
		end
	end
end

function Ailo:InitQuestsTable()
	-- add world bosses to RepeatableQuests table
	for encounterId, questIds in pairs(WorldBosses) do
		local name = EJ_GetEncounterInfo(encounterId)
		RepeatableQuests[name] = { reset = "wb", max = 1, questIds = questIds }
	end

	-- add all quests to TrackedQuests table
	for key, data in pairs(RepeatableQuests) do
		if data.skills then
			for skill, quests in pairs(data.skills) do
				for _, id in ipairs(quests.ids) do
					TrackedQuests[id] = true
				end
			end
		else
			for _, id in ipairs(data.questIds) do
				TrackedQuests[id] = true
			end
		end

		if data.currencies then
			for _, id in ipairs(data.currencies) do
				local info = GetCurrencyInfo(id)
				if info.discovered then
					TrackedCurrencies[id] = { trigger = true, Get = GetCurrencyInfo, tracker = key, name = info.name, total = -1 }
				end
			end
		end
	end
end

--=========================================================================--
--=========================================================================--
--
-- main data acquisition and processing functions
--
--=========================================================================--
--=========================================================================--

TimerMeta = {
	ClearTimer = function(self)
			if self.timer then
				Ailo:CancelTimer(self.timer)
				self.timer = nil
			end
		end,
	ScheduleUpdate = function(self, delay, ...)
			if self.event then
				delay = delay or .5
				if self.timer then
					self:ClearTimer()
					-- reschedule means multiple events happened; drop args so full updates
					self.timer = Ailo:ScheduleTimer(self.updater, delay, self.event)
				else
					self.timer = Ailo:ScheduleTimer(self.updater, delay, self.event, ...)
				end
				self.event = nil
			end
		end,
}
TimerMeta.__index = TimerMeta

Timer = {
	Roster     = setmetatable( { updater = "UpdateRoster" },            TimerMeta ),
	Renown     = setmetatable( { updater = "UpdateRenown" },            TimerMeta ),
	Currencies = setmetatable( { updater = "UpdateCurrencies" },        TimerMeta ),
	Lockouts   = setmetatable( { updater = "TriggerLockoutsUpdate" },   TimerMeta ),
	LFGResets  = setmetatable( { updater = "TriggerLFGResetsUpdate" },  TimerMeta ),
	MythicKey  = setmetatable( { updater = "UpdateMythicPlusKey" },     TimerMeta ),
	MythicPlus = setmetatable( { updater = "UpdateMythicPlus" },        TimerMeta ),
	Bounties   = setmetatable( { updater = "UpdateEmissaryBounties" },  TimerMeta ),
	Callings   = setmetatable( { updater = "TriggerCovenantCallings" }, TimerMeta ),
	Quests     = setmetatable( { updater = "UpdateQuests" },            TimerMeta ),
}

function Ailo:CentralDispatch(event, arg1, ...)
	local idCurrency

	if     event == "PLAYER_LEVEL_UP" or
	       event == "COVENANT_SANCTUM_RENOWN_LEVEL_CHANGED" then
		Timer.Roster.event = event

	elseif event == "MAJOR_FACTION_UNLOCKED" or
	       event == "MAJOR_FACTION_RENOWN_LEVEL_CHANGED" then
		Timer.Renown.event = event

	elseif event == "CURRENCY_DISPLAY_UPDATE" then
		-- check to see if this a currency we track
		local cache = TrackedCurrencies[arg1]
		if cache then
			if RepeatableQuests[cache.tracker] then
				-- a change to this currency means we should update quests as well
				Timer.Quests.event = event
			end
			Timer.Currencies.event = event
			idCurrency = arg1
		end

	elseif event == "CHAT_MSG_SYSTEM" and (tostring(arg1) == strInstanceSaved) then
		-- You are now saved to this instance; refresh Lockout info
		Timer.Lockouts.event = event
	elseif event == "SHOW_LOOT_TOAST" then
		Timer.Quests.event = event
		Timer.Renown.event = event
		Timer.Lockouts.event = event
		Timer.MythicKey.event = event
		Timer.Callings.event = event
		Timer.Bounties.event = event

	elseif event == "LFG_COMPLETION_REWARD" then
		Timer.LFGResets.event = event

	elseif event == "WEEKLY_REWARDS_UPDATE" then
		Timer.MythicPlus.event = event
		Timer.MythicKey.event = event

--	elseif event == "QUEST_LOG_UPDATE" then
--		Timer.Bounties.event = event

	elseif event == "SCENARIO_COMPLETED" or
	      (event == "QUEST_TURNED_IN" and TrackedQuests[arg1]) then
		Timer.Quests.event = event

		-- unfortuneately, some changes to Conquest points are only catchable as quest log changes
		-- REVIEW - is this still true? Or was this for the wacky BFA conquest stuff
		Timer.Currencies.event = event
	elseif (event == "UNIT_QUEST_LOG_CHANGED" and (arg1 == "player")) or
	       (event == "CRAFTINGORDERS_FULFILL_ORDER_RESPONSE" and (arg1 == Enum.CraftingOrderResult.Ok)) then
		Timer.Quests.event = event
	end

	if not InCombatLockdown() then
		-- we will not schedule updates while in combat. once combat ends, we
		-- will get a PLAYER_REGEN_ENABLED message and so end up here where all
		-- timers can be scheduled if there was an previous event for them
		Timer.Roster:ScheduleUpdate()
		Timer.Renown:ScheduleUpdate()
		Timer.Currencies:ScheduleUpdate(nil, idCurrency)
		Timer.Lockouts:ScheduleUpdate()
		Timer.LFGResets:ScheduleUpdate()
		Timer.MythicKey:ScheduleUpdate()
		Timer.MythicPlus:ScheduleUpdate()
		Timer.Bounties:ScheduleUpdate()
		Timer.Callings:ScheduleUpdate()
		Timer.Quests:ScheduleUpdate()
	end
end


function Ailo:UpdatePlayer()
	self:UpdateRoster("UpdatePlayer")

	self:UpdateQuests("UpdatePlayer")
	self:UpdateCurrencies("UpdatePlayer")
	self:UpdateMythicPlusKey("UpdatePlayer")
	self:UpdateMythicPlus("UpdatePlayer")
	self:UpdateEmissaryBounties("UpdatePlayer")

	-- update lockouts for raids and instances by triggering an UPDATE_INSTANCE_INFO event
	self:TriggerLockoutsUpdate("UpdatePlayer")

	-- update resets and LFG rewards by triggering an LFG_UPDATE_RANDOM_INFO event 
	self:TriggerLFGResetsUpdate("UpdatePlayer")

	self:TriggerCovenantCallings("UpdatePlayer")

	-- map checking is for tillers tracker
	self:CheckCurrentMap()

	-- ensure we don't miss a reset
	self.db.global.nextPurge = self:GetNextPurge()
end

function Ailo:UpdateRoster(...)
	local class = select(2,UnitClass("player"))
	charLevel = UnitLevel("player")  -- charLevel is global

	local covenant
	if C_Covenants.GetActiveCovenantID() > 0 then
		covenant = C_CovenantSanctumUI.GetRenownLevel()
		if covenant == 80 then
			covenant = "done"
		end
	end

	local char = { class = class, level = charLevel, renown = { Covenant = covenant } }

	-- find out their two main professions

	-- key is skillLine id for the profession as a whole, df is id for dragonflight version
	-- name is the nonlocalized version, what we will use internally to identify the skill
	local skills = {
	   [171] = { df = 2823, name = "alchemy" },
	   [164] = { df = 2822, name = "blacksmithing" },
	   [333] = { df = 2825, name = "enchanting" },
	   [202] = { df = 2827, name = "engineering" },
	   [182] = { df = 2832, name = "herbalism" },
	   [773] = { df = 2828, name = "inscription" },
	   [755] = { df = 2829, name = "jewelcrafting" },
	   [165] = { df = 2830, name = "leatherworking" },
	   [186] = { df = 2833, name = "mining" },
	   [393] = { df = 2834, name = "skinning" },
	   [197] = { df = 2831, name = "tailoring" },
	}

	local cps = C_ProfSpecs
	local GetChildren = cps.GetChildrenForPath
	local GetState = cps.GetStateForPath

	local function CheckProfession(iProf)
		local idSkill = iProf and select(7, GetProfessionInfo(iProf)) or nil
		local skill = idSkill and skills[idSkill] or nil
		if not skill then return end

		local line = skill.df
		local configID = cps.GetConfigIDForSkillLine(line)

		-- note use of closures; this function references variables set above
		local function TestPath(pathID)
			for _, child in ipairs(GetChildren(pathID)) do
				if not TestPath(child, configID) then return false end
			end
			return (2 == GetState(pathID, configID))
		end

		local fCompleted = true  -- we assume completed until we find the first node that isn't
		local fStarted = false   -- we assume not started until we find the first tab unlocked or unlockable

		for _, tab in ipairs(cps.GetSpecTabIDsForSkillLine(line)) do
			if 0 ~= cps.GetStateForTab(tab, configID) then
				fStarted = true -- at least one tab unlocked or unlockable
			end

			if not TestPath(cps.GetRootPathForTab(tab), configID) then
				fCompleted = false
			end
		end

		if fCompleted then
			char[skill.name] = "completed"
		elseif fStarted then
			char[skill.name] = "started"
		else
			char[skill.name] = nil
		end
	end

	local iProf1, iProf2 = GetProfessions()
	CheckProfession(iProf1)
	CheckProfession(iProf2)

	self.db.global.roster[currentRealmChar] = char

	self:UpdateRenown()
end

function Ailo:UpdateRenown(...)
	local renown = self.db.global.roster[currentRealmChar].renown
	for key, id in pairs(FactionRenown) do
		if cmf.HasMaximumRenown(id) then
			local hasRewardPending = select(4, crep.GetFactionParagonInfo(id))
			renown[key] = hasRewardPending and "reward" or "done"
		else
			local data = cmf.GetMajorFactionData(id)
			renown[key] = data.isUnlocked and data.renownLevel
		end
	end
end

--=========================================================================--
-- Currency Point Trackers
--=========================================================================--

function Ailo:InitCurrenciesTable()
	for tracker, data in pairs(Currencies) do
		local cache = { tracker = tracker, item = data.item, Get = (data.Get and self[data.Get] or GetCurrencyInfo) }
		local info = cache.Get(data.id)

		if data.display == "seasonal" then
			cache.seasonal = true
			cache.seasonMax = (info.maxQuantity and info.maxQuantity > 0) and info.maxQuantity or nil
			cache.earned = -1  -- set the cached total to -1 so first update will set up saved data
		else
			cache.weeklyMax = (info.maxWeeklyQuantity and info.maxWeeklyQuantity > 0) and info.maxWeeklyQuantity or nil
			cache.totalMax = (info.maxQuantity and info.maxQuantity > 0) and info.maxQuantity or nil
			-- set the cached totals to -1 so first update will set up saved data
			cache.weekly = -1
			cache.total  = -1
		end
		cache.name = info.name  -- do we use this anywhere or just as a debug aid?

		TrackedCurrencies[data.id] = cache

		-- if the currency isn't known, then make sure there isn't saved data for it
		-- fixes bug where we switch to a new currency ID for an existing tracker, such
		-- as when the current loot reroll coin changes. if not cleared we'll show the
		-- total from the former currency id
		if not info.discovered then
			local key = strformat(strfDBKey, tracker) -- ie Stormrage.Bob.LootCoins
			self.db.global.currencies["season"][key] = nil
			self.db.global.currencies["weekly"][key] = nil
			self.db.global.currencies["total"][key] = nil
		end
	end
end

-- event hander for CURRENCY_DISPLAY_UPDATE
function Ailo:UpdateCurrencies(event, id, ...)
	local currenciesdb = self.db.global.currencies
	local  totalPoints = currenciesdb.total
	local weeklyPoints = currenciesdb.weekly
	local seasonPoints = currenciesdb.season

	local function UpdateCurrency(id, cache)
		cache = cache or TrackedCurrencies[id]
		if not cache then return false end

		local info = cache.Get(id)
		if not info.discovered then return false end

		-- replace count with count of items in inventory if an item id was supplied
		if cache.item then
			info.quantity = GetItemCount(cache.item, true, nil, true)
		end

		if cache.trigger then
			if cache.total ~= info.quantity then
				-- save and cache the new total amount
				cache.total = info.quantity
			end
		elseif cache.seasonal then
			if cache.total ~= info.quantity or cache.earned ~= info.totalEarned then
				-- save and cache the new total amount
				cache.earned = info.totalEarned
				cache.total = info.quantity

				local key = strformat(strfDBKey, cache.tracker) -- eg Stormrage.Bob.Valor
				if cache.earned > 0 then
					-- we can't save seasonMax because if we log in after it has been increased,
					-- then our saved records would incorrectly say other characters are capped.
					-- instead save the id of the currency so that the cell formatter can find the max.
					seasonPoints[key] = { earned = cache.earned, count = cache.total, id = id }
				else
					seasonPoints[key] = nil
				end
			end
		elseif (cache.total ~= info.quantity or cache.weekly ~= info.quantityEarnedThisWeek) then
			-- may have uncapped by spending some of that currency
			-- local wasCapped = cache.total == cache.totalMax or cache.weekly == cache.weeklyMax

			-- save and cache the new total amount
			cache.total = info.quantity
			cache.weekly = info.quantityEarnedThisWeek

			local key = strformat(strfDBKey, cache.tracker) -- eg Stormrage.Bob.LootCoins

			totalPoints[key] = { max = cache.totalMax, count = cache.total }

			if cache.weeklyMax then -- has a weekly cap
				weeklyPoints[key] = { max = cache.weeklyMax, count = cache.weekly }
			else
				weeklyPoints[key] = nil
			end
		end

		return false
	end

	if event == "CURRENCY_DISPLAY_UPDATE" and type(id) == "number" then
		UpdateCurrency(id)
	else -- if not an update for a specific currency, then do all of them
		for id, cache in pairs(TrackedCurrencies) do
			UpdateCurrency(id, cache)
		end
	end
end


-- DEPRECATED - Dragonflight S1 behavior for currency id 2167
Ailo.GetCatalystChargesInfo = function(id)
	local info = GetCurrencyInfo(id)

	-- currency is not marked as discovered, even when you have it
	info.discovered = true

	-- the currency info doesn't tell us if they gained one this week,
	-- but we should be able to use quest completion to determine that
	info.quantityEarnedThisWeek = IsQuestFlaggedCompleted(72528) and 1 or 0
	info.maxWeeklyQuantity = 1

	return info
end

-- DEPRECATED - BfA specific behavior
-- when completely capped, PVPGetConquestLevelInfo returns 0 for amount, max and quest
-- in that situation, turn the weekly total into 500 so that we'll show capping, but after
-- the weekly reset, we'll have 0/nil for total/weekly totals which is exactly what we want
Ailo.GetConquestInfoBfA = function(conquestID)
	local info = GetCurrencyInfo(conquestID)
	local currentAmount, _, questID = PVPGetConquestLevelInfo();

	if questID == nil or questID == 0 then
		info.quantityEarnedThisWeek = 500
	else
		info.quantityEarnedThisWeek = currentAmount
		TrackedQuests[questID] = true
	end

	info.quantity = currentAmount
	info.maxWeeklyQuantity = 500
	info.maxQuantity = 500

	return info
end


--=========================================================================--
-- Raid and Instance Lockouts
--=========================================================================--

function Ailo:TriggerLockoutsUpdate(...)
	RequestRaidInfo()  -- triggers UPDATE_INSTANCE_INFO event
end

-- event handler for UPDATE_INSTANCE_INFO, process all raid/instance lockouts
function Ailo:UpdateLockouts(...)
	local lockouts = self.db.global.lockouts
	local now = time()
	LFGSearchResultsLockouts = {}
	for iSaved =1, GetNumSavedInstances() do
		local instanceName, _, instanceReset, instanceDifficulty, locked, _, _, isRaid, maxPlayers, difficultyName, maxEncounters, completedEncounters, extendDisabled, instanceId = GetSavedInstanceInfo(iSaved)
		if locked then
			if instanceDifficulty == 7 or instanceDifficulty == 17 then -- LFR instances
				-- we're currently making a large assumption here that the order of the bosses in saved instances info
				-- are grouped and ordered by wing progression. that is, all bosses in a given wing will be together
				-- sequentially and that those groups are in order by wing. So, if wings 1 and 2 have 3 bosses each,
				-- we assume that wing 1 is bosses 1-3 and wing 2 is bosses 4-6.
				-- the data in saved instances we've seen so far supports this and in cases was definitely NOT in the
				-- order the EncounterJournal presents
				local raid = RaidFromInstance[instanceId]
				local iOffset = 0
				for _, wing in ipairs(raid.wings) do
					local killed = 0
					for iBoss = 1+iOffset, iOffset+wing.max do
						killed = killed + (select(3, GetSavedInstanceEncounterInfo(iSaved, iBoss)) and 1 or 0)
					end
					iOffset = iOffset + wing.max

					if killed > 0 then
						local key = strformat(strfDBKey, strformat("%s.%s", raid.name, wing.name)) -- ie Stormrage.Bob.Siege of Orgrimmar.Downfall
						lockouts[key] = { count = killed, max = wing.max, reset = now+instanceReset }
					end
				end
			else
				-- workaround for ulduar bug - Once Freya dies, the 3 Elders in her
				-- room disappear but are still returned as bosses left to kill.
				-- they are returned as uncompleted bosses.
				if instanceId == 603 then
					if select(3, GetSavedInstanceEncounterInfo(iSaved, 13)) then -- if Freya (encounter 13) is dead then
						for elder=10, 12 do -- add 1 to the count for each elder not already being counted
							completedEncounters = completedEncounters + (select(3, GetSavedInstanceEncounterInfo(iSaved, elder)) and 0 or 1)
						end
					end
				end
				-- work around for Assault on Violet Hold bug
				-- 3 of 8 different bosses are randomly chosen,
				-- game reports the bosses you didn't fight as incomplete
				if instanceId == 1544 then
					if completedEncounters == 3 then
						completedEncounters = maxEncounters
					end
				end

				local key = strformat(strfDBKey, strformat("%s.%s", instanceName, instanceTypes[instanceDifficulty])) -- ie Stormrage.Bob.Highmaul.HeroicRaid
				lockouts[key] = { count = completedEncounters, max = maxEncounters, reset = now+instanceReset }

				-- Use this specific format so we can do
				-- if LFGSearchResultsLockouts[btn:GetText()] then ...
				-- later
				LFGSearchResultsLockouts[ strformat("%s (%s)", instanceName, difficultyName) ] = true
			end
		end
	end

	for i=1, GetNumSavedWorldBosses() do
		local bossName, worldBossID, bossReset = GetSavedWorldBossInfo(i)

		-- skip if the id is for a Garrison Invasion
		if worldBossID < 11 or worldBossID > 14 then
			local key = strformat(strfDBKey, strformat("%s.%s", "WorldBosses", bossName)) -- ie Stormrage.Bob.WorldBosses.The Four Celestials
			lockouts[key] = { count = 1, max = 1, reset = now+bossReset }
		end
	end
end

--=========================================================================--
-- LFG Reward Resets
--=========================================================================--

function Ailo:TriggerLFGResetsUpdate(...)
	--[[
	Fires when a random dungeon is completed and the achievement-like
	alert window pops up. The problem is that this DOES NOT update
	the return values of GetLFGDungeonRewards(x), those are updated
	when LFG_UPDATE_RANDOM_INFO is recieved, so force the client to
	call for an update which will be handled by UpdateLFGResets.
	]]--
	RequestLFDPlayerLockInfo()  -- triggers LFG_UPDATE_RANDOM_INFO event
end

-- event handler for LFG_UPDATE_RANDOM_INFO, process all weekly/daily instance resets
function Ailo:UpdateLFGResets(...)
	local dailies  = self.db.global.dailies

	-- random dungeon ids and names can be obtained by:
	-- for i = 1,GetNumRandomDungeons() do print(GetLFGRandomDungeonInfo(i)) end

	local now = time()
	for name, dungeon in pairs(Dungeons) do
		if dungeon.first <= now and now <= dungeon.last and
		   dungeon.minLevel <= charLevel and charLevel <= dungeon.maxLevel and
		   GetLFGDungeonRewards(dungeon.id)
		then
			dailies[strformat(strfDBKey, name)] = { done = true }
		end
	end

	-- daily battleground first win bonus
	-- GetRandomBGInfo() now returns a RandomBgInfo struct (PvpInfoDocumentation.lua)
	-- hasRandomWinToday: Boolean - true if the the player has already won a random battleground once today, false otherwise
	local bgInfo = C_PvP.GetRandomBGInfo()
	if (bgInfo and bgInfo.hasRandomWinToday) then
		dailies[strformat(strfDBKey, "DailyPVP")] = { count = 1, max = 1 }
	end

	local lockouts = self.db.global.lockouts
	local rfreset = self.db.global.nextWeeklyReset
	for _, raid in ipairs(RaidFinder) do
		for _, wing in ipairs(raid.wings) do
			local _, killed = GetLFGDungeonNumEncounters(wing.id)
			if killed > 0 then
				local key = strformat(strfDBKey, strformat("%s.%s", raid.name, wing.name)) -- ie Stormrage.Bob.Siege of Orgrimmar.Downfall
				lockouts[key] = { count = killed, max = wing.max, reset = rfreset }
			end
		end
	end
end

--=========================================================================--
-- Mythic Plus Weekly Best
--=========================================================================--

function Ailo:UpdateMythicPlusKey(event, ...)
	local level = C_MythicPlus.GetOwnedKeystoneLevel()
	local mapID = C_MythicPlus.GetOwnedKeystoneChallengeMapID()
	local name = mapID and C_ChallengeMode.GetMapUIInfo(mapID) or nil

	local dbKey = strformat(strfDBKey, "MythicPlusKey")
	local dbWeeklies = self.db.global.weeklies

	if name then
		dbWeeklies[dbKey] = { level = level, name = name }
	else
		dbWeeklies[dbKey] = nil
	end
end

local fMythicModeMapsAvailable = false

function Ailo:UpdateMythicPlus(event, ...)
	if event == "CHALLENGE_MODE_MAPS_UPDATE" then
		fMythicModeMapsAvailable = true
	elseif not fMythicModeMapsAvailable then
		C_MythicPlus.RequestRewards()
		return
	end

	local activities = C_WeeklyRewards.GetActivities(typeMythicPlus)
	if not typeMythicPlus or not activities or not activities[1] or activities[1].level == 0 then
		return
	end

	local dbWeeklies = self.db.global.weeklies
	local previousThreshold = 0
	for index, activity in ipairs(activities) do
		-- Ailo:Print(index, activity.level, activity.progress, activity.threshold)
		local dbKey = strformat(strfDBKey, "MythicPlusReward"..tostring(index))

		if activity.level > 0 then
			dbWeeklies[dbKey] = { level = activity.level }
		elseif activity.progress > previousThreshold then
			dbWeeklies[dbKey] = { progress = format("%i/%i", activity.progress, activity.threshold) }
		else
			dbWeeklies[dbKey] = nil
		end

		previousThreshold = activity.threshold
	end
end

--=========================================================================--
-- Broken Isles Emissary Bounty caches
--=========================================================================--

function Ailo:UpdateEmissaryBounties(...)
--[===[@debug@
	Ailo:Print("UpdateEmissaryBounties", ...)
--@end-debug@]===]
	local bountydb = self.db.global.bounties
	for bname, b in pairs(bounty_list_bfa) do
		local key = strformat(strfDBKey, bname)

		if IsQuestFlaggedCompleted(b.quest) then
			-- an active bounty, but because its completed we can get no additional information
			bountydb.resets[key] = { done = true }
		else
			local timeLeft = C_TaskQuest.GetQuestTimeLeftMinutes(b.quest) or 0
			if timeLeft > 0 then
				bountydb.timers[bname] = time() + (timeLeft * 60)
			end

			local _, _, _, count, max = GetQuestObjectiveInfo(b.quest, 1, false)
			if count and count > 0 then
				-- they have partial progress towards completion
				bountydb.resets[key] = { count = count, max = max }
			else
				-- not completed and no count? not an active bounty so we can clear it
				bountydb.resets[key] = nil
			end
		end
	end
end

--=========================================================================--
-- Shadowlands Covenant Callings
--=========================================================================--

local function AdjustDay(day, offset)
	-- add or subtract a couple of days
	-- handles the Saturday=7/Sunday=1 transition forwards or backwards
	day = day - 1 -- shift range from 1-7 to 0-6 so modulo arithmetic works right
	day = math.fmod(7 + day + offset, 7)  -- extra 7 allows for subtraction so that we don't modulo a negative number
	day = day + 1 -- shift back
	return day
end

local function GetCurrentCallingDay()
	-- if dr is borked we can't tell what day we're really on, exit
	local dr = GetNextDailyReset()
	if dr == hugeNumber then return nil end

	-- get the day value for right now
	local now = time()
	local today = date("*t", now).wday

	-- if the day value is the same as the daily reset, then we
	-- are actually still part of the previous day, quest wise
	if today == date("*t", dr).wday then
		today = AdjustDay(today, -1)
	end

	return today
end

local function GetCallingDays()
	local today = GetCurrentCallingDay()
	if not today then return nil end

	local days = {}
	for offset = 0, 2 do
		local dayname = CALENDAR_WEEKDAY_NAMES[AdjustDay(today, offset)]
		local calling = "Calling"..tostring(offset+1)
		days[dayname] = calling
		days[calling] = dayname
	end

	return days
end

function Ailo:TriggerCovenantCallings(...)
	C_CovenantCallings.RequestCallings()
end

function Ailo:UpdateCovenantCallings(event, callings)
	-- if they don't have a covenant, then they can't have callings
	-- it will look like they have completed all 3 instead of not having any, just exit
	if C_Covenants.GetActiveCovenantID() == 0 then return end

	local today = GetCurrentCallingDay()
	if not today then return end

	-- init the days as if complete
	local callingsdb = self.db.global.callings
	for offset = 0, 2 do
		local dayname = CALENDAR_WEEKDAY_NAMES[AdjustDay(today, offset)]
		local key = strformat(strfDBKey, dayname)
		callingsdb[key] = { done = true }
	end

	-- now overwrite with ones we know are still active
	for k, info in ipairs(callings) do
		-- if k > 3 then self:Print("calling table has too many entries", k) end
		local timeLeft = C_TaskQuest.GetQuestTimeLeftMinutes(info.questID) or 0
		if timeLeft > 0 then
			local offset = floor(timeLeft / 1440) -- 1440 minutes in a day
			local dayname = CALENDAR_WEEKDAY_NAMES[AdjustDay(today, offset)]
			-- self:Print("Calling", k, dayname, info.questID)
			local key = strformat(strfDBKey, dayname)
			callingsdb[key] = { count = 0, max = 1 }
		else
			-- self:Print("Calling", k, info.questID, time(), "no time left on quest; schedule another request")
			Timer.Callings.event = "retry to correct quest time remaining"
			Timer.Callings:ScheduleUpdate(3)
		end

--		local name = C_QuestLog.GetTitleForQuestID(info.questID)
--		local index = 1 + floor(timeLeft / 1440)
		-- print(id, name, timeLeft, index)
--		ordered[index] = name
	end
end

--=========================================================================--
-- Daily/Weekly Repeatable Quests
--=========================================================================--

function Ailo:UpdateQuests(event, ...)
	local db = self.db.global
	local char = self.db.global.roster[currentRealmChar]
	for objective, data in pairs(RepeatableQuests) do
		local show = false
		local record = {}

		if data.skills then
			for skill,quests in pairs(data.skills) do
				-- the character has a skill in a tracker with quests for that skill
				local state = char[skill]
				if state then show = true end
				if state == "started" then
					local max = quests.max
					if max then
						for _,id in pairs(quests.ids) do
							if IsQuestFlaggedCompleted(id) then
								max = max - 1
							end
						end
						record[skill] = (max > 0) and max or nil
					else
						-- no max, so only 1 quest available. progress assumed to exist as first objective
						record[skill] = 0
						for _,id in pairs(quests.ids) do
							if IsQuestFlaggedCompleted(id) then
								record[skill] = nil
							elseif data.showProgress and GetLogIndexForQuestID(id) then
								local objectives = GetQuestObjectives(id)
								record[skill] = objectives[1].numRequired - objectives[1].numFulfilled
							end
						end
					end
				end
			end

			if next(record) == nil then
				if show then
					record.done = true
				else
					record.count = 0
				end
			end
		else
			local count, max = 0, 0
			for _,id in ipairs(data.questIds) do
				if IsQuestFlaggedCompleted(id) then
					max = max + 1
					count = count + 1
				elseif data.saveIds then
					record[1 + #record] = id
				elseif GetLogIndexForQuestID(id) then
					if data.showProgress then
						-- REVIEW - we're only going to display completion of first objective on assumption
						-- that there is only one we care about. if we want to track progress of quests with
						-- multiple objectives then we'll need to either specify which one (eg showProgress = 2)
						-- or decide how we want to store and display such data
						local objectives = GetQuestObjectives(id)
						max = max + objectives[1].numRequired
						count = count + objectives[1].numFulfilled
						show = data.checklog
					elseif data.checklog then
						max = max + 1
						show = true
					end
				end
			end

			record.count = count
			record.max = (data.showProgress or data.checklog) and max or data.max
		end

		if show or record.count > 0 then
			if data.reset == "w" then  -- w for weekly
				db.weeklies[strformat(strfDBKey, objective)] = record
			elseif data.reset == "b" then  -- b for biweekly
				db.biweeklies[strformat(strfDBKey, objective)] = record
			elseif data.reset == "d" then  -- d for daily
				db.dailies[strformat(strfDBKey, objective)] = record
			elseif data.reset == "wc" then  -- wc for weekly currency
				db.currencies.weekly[strformat(strfDBKey, objective)] = record
			elseif data.reset == "wb" then  -- wb for world boss
				record.reset = GetNextWeeklyReset()
				db.lockouts[strformat(strfDBKey, strformat("%s.%s", "WorldBosses", objective))] = record
			end
		else
			if data.reset == "w" then  -- w for weekly
				db.weeklies[strformat(strfDBKey, objective)] = nil
			elseif data.reset == "b" then  -- b for biweekly
				db.biweeklies[strformat(strfDBKey, objective)] = nil
			elseif data.reset == "d" then  -- d for daily
				db.dailies[strformat(strfDBKey, objective)] = nil
			elseif data.reset == "wc" then  -- wc for weekly currency
				db.currencies.weekly[strformat(strfDBKey, objective)] = nil
			elseif data.reset == "wb" then  -- wb for world boss
				db.lockouts[strformat(strfDBKey, strformat("%s.%s", "WorldBosses", objective))] = nil
			end
		end
	end
end


--=========================================================================--
-- Map Specific Trackers
--=========================================================================--

function Ailo:CheckCurrentMap(...)
	local curMapID = cmp.GetBestMapForUnit("player")

	if curMapID == 376 then -- 376 == valley of the four winds
		self:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED")
	else
		self:UnregisterEvent("UNIT_SPELLCAST_SUCCEEDED")
	end

	-- Hate can be gained without us getting a quest event. Can't use Reputation as players
	-- will be capped with Ve'nari eventually. So, just update every second while in the maw.
	if curMapID == 1543 then -- 1543 == The Maw
		updateJailerTimer = self:ScheduleRepeatingTimer("UpdateJailerEye", 1, "maw polling")
	else
		self:CancelTimer(updateJailerTimer)
		updateJailerTimer = nil
	end
end

--=========================================================================--
-- Eye of the Jailer
--=========================================================================--

function Ailo:UpdateJailerEye(event)
	-- the existence of the repeating timer should let us know
	-- whether or not we're in the maw. if not, do nothing
	if updateJailerTimer then
		-- 2885 appears to be the widget ID used to show Eye of the Jailer progress
		local infoJailerEye = cwm.GetDiscreteProgressStepsVisualizationInfo(2885)

		if infoJailerEye and infoJailerEye.textureKit == "eyeofthejailer" and infoJailerEye.shownState == 1 then
			if infoJailerEye.progressVal == 0 then
				-- print("nilling out jailer eye", cmp.GetBestMapForUnit("player"))
				self.db.global.dailies[strformat(strfDBKey, "JailerEye")] = nil
			else
				local data = self.db.global.dailies[strformat(strfDBKey, "JailerEye")] or { count = 0, max = 5 }
				local progress = floor(infoJailerEye.progressVal / 100) / 10
				if progress ~= data.count then
				--[===[@debug@
					print(format("JailerEye gained %s hate, total=%s (%s)", tostring(progress - data.count), tostring(progress), tostring(infoJailerEye.progressVal)))
				--@end-debug@]===]
					data.count = progress
					self.db.global.dailies[strformat(strfDBKey, "JailerEye")] = data
				end
			end
	--[===[@debug@
		-- if we are in the maw and can't get the widget we should investigate
		else -- if not infoJailerEye or not infoJailerEye.textureKit == "eyeofthejailer" then
			local mapID = cmp.GetBestMapForUnit("player")
			if mapID then
				local mapInfo = cmp.GetMapInfo(mapID)
				-- print("Eye of the Jailer widget id incorrect", mapInfo.mapID, mapInfo.name, event)
			end
	--@end-debug@]===]
		end
	end
end

--=========================================================================--
-- Tillers Farm tracker
--=========================================================================--

function Ailo:UNIT_SPELLCAST_SUCCEEDED(event, unitID, spell, rank, lineID, spellID, ...)
	if unitID == "player" then
		-- self:Print(event, unitID, spell, rank, lineID, spellID, ...)
		if
			130140 == spellID or -- harvest autumn blossom tree
			129674 == spellID or -- harvest fool's cap
			129673 == spellID or -- harvest golden lotus
			111123 == spellID or -- harvest green cabbage
			129687 == spellID or -- harvest green tea leaf
			130025 == spellID or -- harvest jade squash
			123353 == spellID or -- harvest juicycrunch carrot
			129796 == spellID or -- harvest magebulb
			123445 == spellID or -- harvest mogu pumpkin
			123548 == spellID or -- harvest pink turnip
			123355 == spellID or -- harvest plump green cabbage
			130026 == spellID or -- harvest plump jade squash
			123356 == spellID or -- harvest plump juicycruch carrot
			123451 == spellID or -- harvest plump mogu pumpkin
			123549 == spellID or -- harvest plump pink turnip
			123522 == spellID or -- harvest plump red blossom leak
			123380 == spellID or -- harvest plump scallions
			130043 == spellID or -- harvest plump striped melon
			123571 == spellID or -- harvest plump white turnip
			129984 == spellID or -- harvest plump witchberries
			133106 == spellID or -- harvest portal shard
			129705 == spellID or -- harvest rain poppy
			129843 == spellID or -- harvest raptorleaf
			123524 == spellID or -- harvest red blossom leak
			123375 == spellID or -- harvest scallions
			129676 == spellID or -- harvest silkweed
			129757 == spellID or -- harvest snakeroot
			129675 == spellID or -- harvest snow lily
			129887 == spellID or -- harvest songbell
			130168 == spellID or -- harvest spring blossom tree
			130042 == spellID or -- harvest striped melon
			130109 == spellID or -- harvest terrible turnip
			123570 == spellID or -- harvest white turnip
			129814 == spellID or -- harvest windshear cactus
			123516 == spellID or -- harvest winter blossom tree
			129983 == spellID    -- harvest witchberries
		then
			local key = strformat(strfDBKey,"Tillers")
			local crops = tget(self.db.global.dailies,key)
			crops.count = 1 + (crops.count or 0)
			crops.max = 16  -- TODO - REVIEW - match actual size of plot based off progression with tillers
		end
	end
end


--=========================================================================--
-- Mark LFG Search results to see what you are already locked out of.
--=========================================================================--

function Ailo:ColorLFGSearchResults()
	-- Iterate over LFGClist search results and mark everything you are already
	-- locked out of!
	local profile = self.db.profile
	local index = 1
	local btn   = _G["LFGListSearchPanelScrollFrameButton"..index]
	while btn do
		local act =  btn.ActivityName:GetText()
		if LFGSearchResultsLockouts[act]then
			btn.ActivityName:SetTextColor(profile.savedraid.r, profile.savedraid.g, profile.savedraid.b, profile.savedraid.a)
		end
		index = index +1

		btn = _G["LFGListSearchPanelScrollFrameButton"..index]
	end
end


local function hook_LFGListScroll()
	Ailo:ColorLFGSearchResults()
end

-- Hook the OnVerticalScroll of the ScrollFrame

-- hooksecurefunc(LFGListSearchPanelScrollFrame, "SetVerticalScroll", hook_LFGListScroll )

-- Hook every the button's OnClick
do
	local index = 1
	local btn   = _G["LFGListSearchPanelScrollFrameButton"..index]
	while btn do
		btn:HookScript("OnClick", hook_LFGListScroll)
		index = index +1
		btn = _G["LFGListSearchPanelScrollFrameButton"..index]
	end
end

--=========================================================================--
-- Mark LFR Dropdown
--=========================================================================--

-- This replaces the DropDown setup code for the LFR Frame with our own
-- version. When a new build deploys, check if this function changed, if so,
-- update and set lastCheckedWowBuild to the right value if everything works

local lastCheckedWowBuild = 24461
local currentWowBuild = tonumber((select(2,GetBuildInfo())))
if currentWowBuild <= lastCheckedWowBuild then
	local function isRaidFinderDungeonDisplayable(id)
		local name, typeID, subtypeID, minLevel, maxLevel, _, _, _, expansionLevel = GetLFGDungeonInfo(id);
		local myLevel = UnitLevel("player");
		return myLevel >= minLevel and myLevel <= maxLevel and EXPANSION_LEVEL >= expansionLevel;
	end

	function RaidFinderQueueFrameSelectionDropDown_Initialize(self)

	   local sortedDungeons = { };
	   local function InsertDungeonData(id, name, mapName, isAvailable)
		  local t = { id = id, name = name, mapName = mapName, isAvailable = isAvailable };
		  local foundMap = false;
		  for i = 1, #sortedDungeons do
			 if ( sortedDungeons[i].mapName == mapName ) then
				foundMap = true;
			 else
				if ( foundMap ) then
				   tinsert(sortedDungeons, i, t);
				   return;
				end
			 end
		  end
		  tinsert(sortedDungeons, t);
	   end

	   -- If we ever change this logic, we also need to change the logic in RaidFinderFrame_UpdateAvailability
	   for i=1, GetNumRFDungeons() do
		  local dungeonInfo = { GetRFDungeonInfo(i) };
		  local id = dungeonInfo[1];
		  local name = dungeonInfo[2];
		  local mapName = dungeonInfo[20];
		  local isAvailable, isAvailableToPlayer, hideIfUnmet = IsLFGDungeonJoinable(id);
		  if( not hideIfUnmet or isAvailable ) then
			 if ( isAvailable or isAvailableToPlayer or isRaidFinderDungeonDisplayable(id) ) then
				InsertDungeonData(id, name, mapName, isAvailable);
			 end
		  end
	   end

	   local info = UIDropDownMenu_CreateInfo();
	   local currentMapName = nil;
	   for i = 1, #sortedDungeons do
		  if ( currentMapName ~= sortedDungeons[i].mapName ) then
			 currentMapName = sortedDungeons[i].mapName;
			 info.text = sortedDungeons[i].mapName;
			 info.isTitle = 1;
			 info.notCheckable = 1;
			 info.tooltipOnButton = nil;
			 UIDropDownMenu_AddButton(info);
			 info.notCheckable = nil;
		  end
		  if ( sortedDungeons[i].isAvailable ) then
			 info.text = sortedDungeons[i].name; --Note that the dropdown text may be manually changed in RaidFinderQueueFrame_SetRaid
			 info.value = sortedDungeons[i].id;
			 info.isTitle = nil;
			 info.func = RaidFinderQueueFrameSelectionDropDownButton_OnClick;
			 info.disabled = nil;
			 info.checked = (RaidFinderQueueFrame.raid == info.value);
			 info.tooltipWhileDisabled = nil;
			 info.tooltipOnButton = 1;
			 info.tooltipTitle = RAID_BOSSES;
			 local encounters;
			 local numEncounters = GetLFGDungeonNumEncounters(sortedDungeons[i].id);
			 local alive = 0
			 for j = 1, numEncounters do
				local bossName, _, isKilled = GetLFGDungeonEncounterInfo(sortedDungeons[i].id, j);
				local colorCode = "";
				if ( isKilled ) then
				   colorCode = RED_FONT_COLOR_CODE;
				else
				   alive = alive + 1
				end
				if encounters then
				   encounters = encounters.."|n"..colorCode..bossName..FONT_COLOR_CODE_CLOSE;
				else
				   encounters = colorCode..bossName..FONT_COLOR_CODE_CLOSE;
				end
			 end
			 if alive == 0 then
				info.text = RED_FONT_COLOR_CODE .. info.text ..FONT_COLOR_CODE_CLOSE
			 end

			 info.tooltipText = encounters;
			 UIDropDownMenu_AddButton(info);
		  else
			 info.text = sortedDungeons[i].name; --Note that the dropdown text may be manually changed in RaidFinderQueueFrame_SetRaid
			 info.value = sortedDungeons[i].id;
			 info.isTitle = nil;
			 info.func = nil;
			 info.disabled = 1;
			 info.checked = nil;
			 info.tooltipWhileDisabled = 1;
			 info.tooltipOnButton = 1;
			 info.tooltipTitle = YOU_MAY_NOT_QUEUE_FOR_THIS;
			 info.tooltipText = LFGConstructDeclinedMessage(sortedDungeons[i].id);
			 UIDropDownMenu_AddButton(info);
		  end
	   end
	end
end

--=========================================================================--
--=========================================================================--
--
-- saved data maintenance functions
--
--=========================================================================--
--=========================================================================--

function Ailo:PurgeExpiredData()
	local globaldb = self.db.global
	local now = time()

	if now > globaldb.nextWeeklyReset then
		globaldb.currencies.weekly = {}
		globaldb.weeklies = {}
		local wr = GetNextWeeklyReset()
		globaldb.nextWeeklyReset = wr
		globaldb.nextWeeklyResetDate = date(nil, wr)
	end

	if now > globaldb.nextBiweeklyReset then
		globaldb.biweeklies = {}
		local br = GetNextBiweeklyReset()
		globaldb.nextBiweeklyReset = br
		globaldb.nextBiweeklyResetDate = date(nil, br)
	end

	if now > globaldb.nextDailyReset then
		globaldb.dailies = {}
		local dr = GetNextDailyReset()
		globaldb.nextDailyReset = dr
		globaldb.nextDailyResetDate = date(nil, dr)
	end

	local lockouts = globaldb.lockouts
	for key, data in pairs(lockouts) do
		if now > data.reset then
			lockouts[key] = nil
		end
	end

	local days = GetCallingDays()
	if days then
		local callings = globaldb.callings
		for key, data in pairs(callings) do
			local _, _, _, day = breakKey(key)
			if not days[day] then
				callings[key] = nil
			end
		end
	end
end

function Ailo:GetNextPurge()
	local globaldb = self.db.global
	local nextPurge = min(     GetNextDailyReset(),      GetNextBiweeklyReset(),     GetNextWeeklyReset(),
	                      globaldb.nextDailyReset, globaldb.nextBiweeklyReset, globaldb.nextWeeklyReset)

	for _, data in pairs(globaldb.lockouts) do
		nextPurge = min(nextPurge, data.reset)
	end

	return nextPurge
end

function Ailo:WipeDB()
	local globaldb = self.db.global
	globaldb.nextPurge = min(GetNextDailyReset(), GetNextBiweeklyReset(), GetNextWeeklyReset())
	globaldb.nextDailyReset  = 0
	globaldb.nextWeeklyReset = 0
	globaldb.roster = {}
	globaldb.currencies = { total = {}, weekly = {}, season = {} }
	globaldb.dailies    = {}
	globaldb.biweeklies = {}
	globaldb.weeklies   = {}
	globaldb.bounties   = { resets = {}, timers = {} }
	globaldb.callings   = {}
	globaldb.lockouts   = {}
end


function Ailo:ManualPlayerUpdate()
	self:Output(L["Updating data for current player."])

	local globaldb = self.db.global
	globaldb.roster[currentRealmChar] = nil

	local subtables = { "dailies", "biweeklies", "weeklies", "lockouts", "callings" }
	for _, subname in pairs(subtables) do
		local subtable = globaldb[subname]
		for key in pairs(subtable) do
			if strfind(key, currentRealmChar, 1, true) then
				subtable[key] = nil
			end
		end
	end

	local subtables = { "total", "weekly", "season" }
	for _, subname in pairs(subtables) do
		local subtable = globaldb.currencies[subname]
		for key in pairs(subtable) do
			if strfind(key, currentRealmChar, 1, true) then
				subtable[key] = nil
			end
		end
	end

	local subtable = globaldb.bounties.resets
	for key in pairs(subtable) do
		if strfind(key, currentRealmChar, 1, true) then
			subtable[key] = nil
		end
	end

	self:UpdatePlayer()
end


function Ailo:GetDifficultyFormatString(type)
	local difficulty = instanceTypeDifficulty[type]
	if difficulty then
		local abbr = self.db.profile.difficultyAbbr[difficulty]
		if abbr then
			return strformat("%s  %%s", abbr)
		end
	end
	return "%s"
end

function Ailo:GetDifficultyAbbr(type)
	local difficulty = instanceTypeDifficulty[type]
	if difficulty then
		local abbr = self.db.profile.difficultyAbbr[difficulty]
		if abbr then
			return abbr
		end
	end
	return difficulty or type
end

function Ailo:GetInstanceAbbr(instanceName)
	if not self.db.profile.instanceAbbr[instanceName] then
		-- Has no abbreviation yet, try it with a somewhat good guess
		-- Tries to get the first char of every word, does not go well with utf-8 chars
		self.db.profile.instanceAbbr[instanceName] = AbbreviateString(instanceName)
	end

	return ( self.db.profile.instanceAbbr[instanceName] ~= "" and self.db.profile.instanceAbbr[instanceName] or nil )
end


function Ailo:SetupClasscoloredFonts()
	local class, color, CHOSEN_CLASS_COLORS

	if self.db.profile.useCustomClassColors and CUSTOM_CLASS_COLORS then
		CHOSEN_CLASS_COLORS = CUSTOM_CLASS_COLORS
	else
		CHOSEN_CLASS_COLORS = RAID_CLASS_COLORS
	end
	for class,color in pairs(CHOSEN_CLASS_COLORS) do
		if not RAID_CLASS_COLORS_FONTS[class] then 
			RAID_CLASS_COLORS_FONTS[class] = CreateFont("ClassFont"..class)
			RAID_CLASS_COLORS_FONTS[class]:CopyFontObject(_G.GameTooltipText)
		end
		RAID_CLASS_COLORS_FONTS[class]:SetTextColor(color.r, color.g, color.b)
	end
end


--=========================================================================--
--=========================================================================--
--
-- Tooltip and Options Display
--
--=========================================================================--
--=========================================================================--

--=========================================================================--
-- Cell Format Functions
--=========================================================================--

-- WARNING - always set a cell value before setting color

function CellFormat:Default(tooltip, row, col, data)
	local profile = Ailo.db.profile
	local color = profile.freeraid
	local value = ""

	if data then
		if data.done or data.count == data.max then
			color = profile.savedraid
		elseif profile.showLeft and data.count > 0 then
			value = data.max - data.count
		end
	end

	tooltip:SetCell(row, col, value)
	tooltip:SetCellColor(row, col, color.r, color.g, color.b, color.a)
end

function CellFormat:DefaultEmpty(tooltip, row, col, data)
	-- same as default, except if there is no data then cell is truly empty (not green)
	if data then
		local profile = Ailo.db.profile
		local color = profile.freeraid
		local value = ""

		if data.done or data.count == data.max then
			color = profile.savedraid
		elseif profile.showLeft and data.count > 0 then
			value = data.max - data.count
		end

		tooltip:SetCell(row, col, value)
		tooltip:SetCellColor(row, col, color.r, color.g, color.b, color.a)
	end
end

function CellFormat:SimpleCap(tooltip, row, col, data)
	local profile = Ailo.db.profile
	local color = profile.freeraid
	local value = ""

	if data then
		if data.done or data.count == data.max then
			color = profile.savedraid
		elseif data.count > 0 then
			value = data.count
		end
	end

	tooltip:SetCell(row, col, value)
	tooltip:SetCellColor(row, col, color.r, color.g, color.b, color.a)
end

function CellFormat:Renown(tooltip, row, col, data)
	if data then
		local profile = Ailo.db.profile
		local color = profile.freeraid
		local value = ""

		if data.done then
			color = profile.savedraid
		end

		if data.reward then
			value = strformat("|T%s:0|t", 133641)
		elseif data.count and data.count > 0 then
			value = data.count
		end

		tooltip:SetCell(row, col, value)
		tooltip:SetCellColor(row, col, color.r, color.g, color.b, color.a)
	end
end

-- nonlocalized skill names in alphabetical order
local skillOrder = { "alchemy", "blacksmithing", "enchanting", "engineering", "herbalism", "inscription", "jewelcrafting", "leatherworking", "mining", "skinning", "tailoring" }
function CellFormat:ProfessionKnowledge(tooltip, row, col, data)
	if data then
		local profile = Ailo.db.profile
		local color = profile.freeraid
		local value = ""

		if data.done then
			color = profile.savedraid
		else
			for _,skill in ipairs(skillOrder) do
				local remaining = data[skill]
				if type(remaining) == "number" then
					value = strformat("%s%s|TInterface\\Icons\\ui_profession_%s:0|t", value, (remaining == 0 and "" or ("  " .. tostring(remaining))), skill)
				end
			end
			-- remove leading white space if it exists
			if strsub(value,1, 2) == "  " then value = strsub(value, 3) end
		end

		tooltip:SetCell(row, col, value, CellFonts.iconTable)
		tooltip:SetCellColor(row, col, color.r, color.g, color.b, color.a)
	end
end

function CellFormat:ChestQuality(tooltip, row, col, data)
	tooltip:SetCell(row, col, "")

	if not data or data.count == 0 then
		tooltip:SetCellColor(row, col, .64, .21, .93, 1)  -- epic purple
	elseif data.count == 1 then
		tooltip:SetCellColor(row, col,   0, .44, .87, 1)  -- rare blue
	elseif data.count == 2 then
		tooltip:SetCellColor(row, col, .12,   1,   0, 1)  -- uncommon green
	elseif data.count > 2 then
		tooltip:SetCellColor(row, col,   1,   1,   1, 1)  -- common white
	end
end

function CellFormat:RemainingIcons(tooltip, row, col, data)
	local profile = Ailo.db.profile
	local color = profile.freeraid
	local value = ""

	if data then
		if data.done or data.count == data.max then
			color = profile.savedraid
		else
			for _, id in ipairs(data) do
				local icon = icons[id]
				if icon then
					value = strformat("%s|T%s:0|t", value, icon)
				end
			end
		end
	end

	tooltip:SetCell(row, col, value, CellFonts.iconTable)
	tooltip:SetCellColor(row, col, color.r, color.g, color.b, color.a)
end

function CellFormat:MythicPlusReward(tooltip, row, col, data)
	local profile = Ailo.db.profile
	local color = profile.freeraid
	local value = ""

	if data then
		if data.level then
			value = data.level
			if value >= maxMythicPlus then
				color = profile.savedraid
			end
		elseif data.progress then
			value = data.progress
		end
	end

	tooltip:SetCell(row, col, value)
	tooltip:SetCellColor(row, col, color.r, color.g, color.b, color.a)
end

function CellFormat:MythicPlusKey(tooltip, row, col, data)
	local profile = Ailo.db.profile
	local color = profile.freeraid
	local value = ""

	if data and data.level and data.name then
		value = format("%s %s", tostring(data.level), Ailo:GetInstanceAbbr(data.name))
	end

	tooltip:SetCell(row, col, value)
	tooltip:SetCellColor(row, col, color.r, color.g, color.b, color.a)
end

function CellFormat:CurrencySingle(tooltip, row, col, data)
	local profile = Ailo.db.profile
	local color = profile.freeraid
	local value = ""

	if data then
		value = data.count
		if data.max and data.count >= data.max then
			color = profile.savedraid
		end
	end

	tooltip:SetCell(row, col, value)
	tooltip:SetCellColor(row, col, color.r, color.g, color.b, color.a)
end

function CellFormat:CurrencyDouble(tooltip, row, col, data)
	local profile = Ailo.db.profile
	local color = profile.freeraid
	local value = ""

	if data then
		value = strformat("%i  %i", data.total.count, data.weekly.count)
		if data.total.count >= data.total.max or data.weekly.count >= data.weekly.max then
			color = profile.savedraid
		end
	end

	tooltip:SetCell(row, col, value)
	tooltip:SetCellColor(row, col, color.r, color.g, color.b, color.a)
end

function CellFormat:CurrencySeasonal(tooltip, row, col, data)
	local profile = Ailo.db.profile
	local color = profile.freeraid
	local value = ""

	if data then
		local cache = data.id and TrackedCurrencies[data.id] or nil
		local seasonMax = cache and cache.seasonMax

		-- seasonal currencies can switch to uncapped near end of season
		if seasonMax then
			local total = data.earned - seasonMax
			if total == 0 then
				-- hit cap, color cell red
				color = profile.savedraid
			else
				-- show the amount they need to cap
				value = total
			end
		else
			-- if no season cap, then just show the current total
			value = data.count
		end
	end

	tooltip:SetCell(row, col, value)
	tooltip:SetCellColor(row, col, color.r, color.g, color.b, color.a)
end

function CellFormat:DualCappedCurrency(tooltip, row, col, data)
	local profile = Ailo.db.profile
	local color = profile.freeraid
	local value = ""

	if data then
		local total = data.total and data.total.count or 0
		local totalLeft = (data.total and data.total.max or hugeNumber) - total

		local weekly = data.weekly and data.weekly.count or 0
		local weeklyLeft = (data.weekly and data.weekly.max or hugeNumber) - weekly

		if totalLeft == 0 then
			-- hit total cap, color cell red and show total amount so user can see this is different from weekly capping
			value = total
			color = profile.savedraid
		elseif weeklyLeft == 0 then
			-- hit weekly cap, color cell red but show no number; user is in a good place
			color = profile.savedraid
		elseif weeklyLeft >= totalLeft and total ~= 0 then
			-- user will hit total cap before weekly cap, color cell green and show total amount
			value = total
		elseif weekly > 0 then
			-- will hit weekly cap before total, color cell green and show amount earned this week
			value = weekly
		end
	end

	tooltip:SetCell(row, col, value)
	tooltip:SetCellColor(row, col, color.r, color.g, color.b, color.a)
end

-- This currency sucks because there is no way to distinguish between not being capped because
-- you've spent charges and not being capped because you could still earn more.
-- Show charges and color red if they have none.
function CellFormat:CatalystCharges(tooltip, row, col, data)
	if data then
		local profile = Ailo.db.profile
		local color = profile.freeraid
		local value = ""

		local total = data.total and data.total.count or 0
		local max = TrackedCurrencies[2167].totalMax
		if total > 0 then
			value = total
		end

		local weekly = data.weekly and data.weekly.count or 0
		local weeklyLeft = (data.weekly and data.weekly.max or hugeNumber) - weekly
		if weeklyLeft == 0 or total == max then
			color = profile.savedraid
		end

		tooltip:SetCell(row, col, value)
		tooltip:SetCellColor(row, col, color.r, color.g, color.b, color.a)
	end
end

function CellFormat:ResearchersUnderFire(tooltip, row, col, data)
	local profile = Ailo.db.profile
	local color = profile.freeraid
	local value = ""

	if data then
		local incomplete = {}
		for _, id in ipairs(data) do
			incomplete[id - 75626] = true
		end

		if not incomplete[4] then
			color = profile.savedraid
		elseif not incomplete[3] then
			value = L["Tier 3"]
		elseif not incomplete[2] then
			value = L["Tier 2"]
		elseif not incomplete[1] then
			value = L["Tier 1"]
		end
	end

	tooltip:SetCell(row, col, value)
	tooltip:SetCellColor(row, col, color.r, color.g, color.b, color.a)
end



--=========================================================================--
-- Tooltip Filtering Functions
--=========================================================================--

local function PrepRosterSlice()
	-- returns a roster filtered by level check into which resets can be added
	local minLevel = Ailo.db.profile.minLevel
	local roster = {}

	for realmChar, data in pairs(Ailo.db.global.roster) do
		if minLevel <= data.level then
			local realm, char = strmatch(realmChar, "(.+)%.(.+)")
			roster[realmChar] = { name = char, level = data.level, color = RAID_CLASS_COLORS_FONTS[data.class], resets = {} }
		end
	end

	-- structure is { realmChar = { (char)name, level, color, resets = { name = data } } }
	-- eg count = roster["Stormrage.Bob"].resets.LootCoins.count
	return roster
end

local function SortRosterSlice(roster)
	-- roster param structure is { realmChar = { (char)name, level, color, resets = { name = data } } }
	-- basically, turn the key into sorted tables which point to the value
	-- TODO - we can offer additional methods of sorting and here's where they would kick
	-- TODO - sort by character level, iLevel, etc
	local sortedRoster = {}
	local temp = {}
	for realmChar,data in pairs(roster) do
		tinsert(temp, realmChar)
	end

	local currentRealm = breakKey(currentRealmChar)
	local charSort = function(a,b)
		-- current character should be first one displayed
		if a == currentRealmChar then
			return true
		elseif b == currentRealmChar then
			return false
		end

		local aRealm = breakKey(a)
		local bRealm = breakKey(b)

		-- if they are same realm, compare names
		if aRealm == bRealm then
			return a < b
		-- if not same and one is current realm, it goes first
		elseif aRealm == currentRealm then
			return true
		elseif bRealm == currentRealm then
			return false
		-- otherwise, compare server names
		else
			return aRealm < bRealm
		end
	end

	tsort(temp, charSort)

	local curRealm
	for _,realmChar in ipairs(temp) do
		local realm, char = strmatch(realmChar, "(.+)%.(.+)")

		if curRealm ~= realm then
			curRealm = realm
			tinsert(sortedRoster, { name = realm, chars = {} })
		end

		tinsert( sortedRoster[#sortedRoster].chars, roster[realmChar] )
	end

	-- structure is { [iRealm] = { name, chars = { [iChar] = { name, level, color, resets = { name = data } } } } }
	-- eg count = sortedRoster[1].chars[1].resets.LootCoins.count
	return sortedRoster
end


--=========================================================================--
-- PrepareTooltip - build and format the tooltip to show lockout state
--=========================================================================--

function Ailo:PrepareTooltip(tooltip)


	--[[ Cell are just colored green/red
	       Coins Daily    Bar
		   T   W  Foo  H   N  LFR
	Char1 [ ] [ ] [ ] [ ] [ ] [ ] [ ]
	Char2 [ ] [ ] [ ] [ ] [ ] [ ] [ ]
	Char3 [ ] [ ] [ ] [ ] [ ] [ ] [ ]
	]]--


	local globaldb = self.db.global
	local profiledb = self.db.profile

	if time() > globaldb.nextPurge then
		self:PurgeExpiredData()
		globaldb.nextPurge = self:GetNextPurge()
	end

	-- we don't always get the events we should be getting for quests when they are completed
	-- if we're going to show the tooltip lets do another pass through quests in case one has changed
	-- self:UpdateQuests("Showing Tooltip")

	local Columns = {}  -- keyed by resetname and colindex, value is { reset, col, Format }

	local function MyAddColumn(nameType, displayName, formatFunc, font)
		if nameType then
			if not displayName then
				-- print("MyAddColumn: missing displayName for", nameType)
				displayName = nameType
			end
			if not formatFunc then
				formatFunc = CellFormat.Default
			end

			if not Columns[nameType] then
				local colindex = tooltip:AddColumn("CENTER")
				tooltip:SetCell(2, colindex, displayName, font)

				local data = { reset = nameType, col = colindex, Format = formatFunc }
				Columns[nameType] = data
				Columns[colindex] = data
				-- print("adding column", colindex, nameType, displayName)
			else
				--print("a column already exists for", nameType)
			end
		end
	end


	-- REVIEW - hard to see right now what columns a spanner covers, we should color the cell in
	-- some way to distinguish it from the background. One idea was to have a set of colors with
	-- each spanner using the next color in the set and repeating as necessary. For example, the
	-- spanners could be colored cyan, yellow, magenta, cyan, yellow, etc. My palette picking is
	-- not up to par, so for now we'll just use a dark gray.
	local spannerColors = {
		-- { r = 255, g = 0, b = 0, a = 1 },
		-- { r = 255, g = 128, b = 0, a = 1 },
		{ r = 255, g = 255, b = 0, a = 1 },
		-- { r = 0, g = 255, b = 0, a = 1 },
		{ r = 0, g = 255, b = 255, a = 1 },
		-- { r = 0, g = 0, b = 255, a = 1 },
		{ r = 127, g = 0, b = 255, a = 1 },
		{ r = 50, g = 50, b = 50, a = 1 },
	}
	setmetatable( spannerColors, { __index = function(t,i) return t[ fmod(i-1, #t) + 1 ] end } )
	local numSpanners = 1

	local function AddSpanner(first, last, fullname, abbr, font)
		-- TODO - can we, from the size of span and maybe the character count of fullname,
		-- make a guess as to whether we could use the full name instead of Abbr?
		if first <= last then
			tooltip:SetCell(1, first, abbr, font, nil, 1 + last - first)

			-- numSpanners = numSpanners + 1
			-- local color = spannerColors[numSpanners]
			-- tooltip:SetCellColor(1, first, color.r, color.g, color.b, color.a)
			tooltip:SetCellColor(1, first, 50, 50, 50, 1)
		end
	end


	-- PrintGlobalRoster()
	local roster = PrepRosterSlice()  -- structure is { realmChar = { (char)name, level, color, resets = { name = data } } }
	-- PrintRosterSlice(roster)

	local spanStart
	tooltip:AddHeader()
	tooltip:AddHeader()
	-- tooltip:AddColumn()


	--==============================================================--
	--  CURRENCIES, DAILIES AND WEEKLIES
	--
	--  add the columns defined in the Trackers table
	--==============================================================--

	local found = {}  -- column keynames with saved data that can be displayed

	
	do  -- add renown columns
		-- two pass system, first identifies which columns will be shown. second handles data.
		-- this is so we can hide a column if everyone is capped unless someone has a reward
		local rosterdb = globaldb.roster
		for realmChar, data in pairs(roster) do
			local renown = rosterdb[realmChar].renown or {}
			for name, value in pairs(renown) do
				if profiledb.show[name] and value and value ~= "done" then
					found[name] = true
				end
			end
		end
		for realmChar, data in pairs(roster) do
			local renown = rosterdb[realmChar].renown or {}
			for name, value in pairs(renown) do
				if found[name] and value then
					if value == "reward" then
						data.resets[name] = { done = true, reward = true }
					elseif value == "done" then
						data.resets[name] = { done = true }
					else
						data.resets[name] = { count = value }
					end
				end
			end
		end
	end

	-- TODO - currencies need another look at for when we should or shouldn't display them
	-- if the only reason we're displaying a row is to show you have 0 in a currency, that 
	-- would be kinda wrong. besides 0, when else would we maybe not want to show it?
	for branch, branchdb in pairs(globaldb.currencies) do
		for dbkey,data in pairs(branchdb) do
			local realm, char, realmChar, name = breakKey(dbkey)

			if roster[realmChar] and profiledb.show[name] then
				local currency = Currencies[name]
				local display = currency and currency.display or ""
				if display == "seperate" then
					local nameType = strformat("%s.%s", name, branch)
					roster[realmChar].resets[nameType] = data
					found[nameType] = true
					-- print("found", nameType)
				elseif display == "combined" then
					local temp = roster[realmChar].resets[name] or {}
					temp[branch] = data
					roster[realmChar].resets[name] = temp
					found[name] = true
					-- print("found", name, branch)
				elseif display == "seasonal" then
					roster[realmChar].resets[name] = data
					found[name] = true
					-- print("found", name)
				end
			end
		end
	end


	local function DoQuests(db)
		for dbkey,data in pairs(db) do
			local realm, char, realmChar, name, _, nameType = breakKey(dbkey)

			if roster[realmChar] and profiledb.show[name] then
				roster[realmChar].resets[nameType] = data
				found[nameType] = true
				-- print("found", nameType)
			end
		end
	end
	DoQuests(globaldb.dailies)
	DoQuests(globaldb.weeklies)
	DoQuests(globaldb.biweeklies)


	-- add columns for found currencies, dailies and weeklies
	for _, span in ipairs(OrderedColumns) do
		spanStart = 1 + tooltip:GetColumnCount()
		for _, column in ipairs(span.columns) do
			if column.key then
				if column.showEmpty then
					-- print("showEmpty", column.key, breakKey(column.key), profiledb.show[breakKey(column.key)])
					found[column.key] = profiledb.show[breakKey(column.key)]
				end
				if found[column.key] then
					MyAddColumn(column.key, column.abbr, column.format, column.font)
				end
			end
		end
		AddSpanner(spanStart, tooltip:GetColumnCount(), span.fullname, span.abbr, span.font)
	end


	--==============================================================--
	-- BOUNTIES, EMMISSARIES AND CALLINGS
	--
	-- add the columns for the 3 day groups of objectives
	--==============================================================--

	if profiledb.show.Callings then
		local days = GetCallingDays()
		if days then
			for dbkey,data in pairs(globaldb.callings) do
				local realm, char, realmChar, day = breakKey(dbkey)
				if not roster[realmChar] then
					-- print(strformat("%s skipped because %s is not in filtered roster", nameType, realmChar))
				elseif not days[day] then
					-- print(strformat("$s skipped because day is out of range", dbkey))
				else
					roster[realmChar].resets[days[day]] = data
				end
			end
		end

		spanStart = 1 + tooltip:GetColumnCount()
		MyAddColumn("Calling1", profiledb.columnName[ days["Calling1"] ])
		MyAddColumn("Calling2", profiledb.columnName[ days["Calling2"] ])
		MyAddColumn("Calling3", profiledb.columnName[ days["Calling3"] ])
		AddSpanner(spanStart, tooltip:GetColumnCount(), L["Callings"], profiledb.columnName.Callings)
	end


	if profiledb.show.Bounties then
	-- Bounties / Emissary Caches
	-- Determine active ones
	-- For a few minutes after the time the oldest bounty gets replaced, the tooltip may show 4 bounties
	-- IsQuestFlaggedCompleted returns true even though on the emissary quest that is already out
	-- C_TaskQuest.GetQuestTimeLeftMinutes() sometimes returns nil
		local timers = Ailo.db.global.bounties.timers
		local active_bounties = {}

		for bname, b in pairs(bounty_list_bfa) do
			if timers[bname] > time() or (C_TaskQuest.GetQuestTimeLeftMinutes(b.quest) or 0) > 0 or
			   IsQuestFlaggedCompleted(b.quest) or GetQuestLink(b.quest) then
				-- print(strformat("active_bounties[%s]", bname))
				active_bounties[bname] = true
				active_bounties[#active_bounties+1] = bname
			end
		end
		tsort( active_bounties, function(a,b) return timers[a] < timers[b] end )

		for dbkey,data in pairs(globaldb.bounties.resets) do
			local realm, char, realmChar, bname = breakKey(dbkey)
			if not roster[realmChar] then
				-- print(strformat("%s skipped because %s is not in filtered roster", nameType, realmChar))
			elseif not active_bounties[bname] then
				-- print(strformat("not active_bounties[%s]", bname))
			else
				roster[realmChar].resets[strformat("Bounty.%s", bname)] = data
			end
		end
	
		spanStart = 1 + tooltip:GetColumnCount()
		for _, bname in ipairs(active_bounties) do
			MyAddColumn(strformat("Bounty.%s", bname), strformat("|T%s:0|t", bounty_list_bfa[bname].icon), nil, CellFonts.iconHeader) -- abbr name actually, if abbr name doesn't exist we won't add
		end
		AddSpanner(spanStart, tooltip:GetColumnCount(), L["Bounties"], profiledb.columnName.Bounties)
	end


	--==============================================================--
	-- DUNGEON, RAID AND WORLD BOSS LOCKOUTS
	--==============================================================--

	local Lockouts = {}
	local WorldBosses = {}  -- = { [i] = name }
	for dbkey,data in pairs(globaldb.lockouts) do
		local realm, char, realmChar, name, type, nameType = breakKey(dbkey)
		if not roster[realmChar] then
			-- print(strformat("%s skipped because %s is not in filtered roster", nameType, realmChar))
		elseif name == "WorldBosses" then
			if not profiledb.show.WorldBosses then
				-- print(strformat("%s skipped because profile setting for %s is %s", dbkey, "WorldBosses", tostring(profiledb.show.WorldBosses)))
			else
				roster[realmChar].resets[nameType] = data
				WorldBosses[1 + #WorldBosses] = type
			end
		elseif not profiledb.show["5Man"] and strmatch(type, "%a+Dungeon") then
			-- print(strformat("%s skipped because profile setting for %s is %s", dbkey, "5Man", tostring(profiledb.show["5Man"])))
		else
			roster[realmChar].resets[nameType] = data
			tget(Lockouts,name)[type] = nameType
			-- example: Lockouts = { "Siege of Orgrimmar" = { "HeroicRaid" = "Siege of Orgrimmar.Mythic", "Downfall" = "Siege of Orgrimmar.Downfall" } }
		end

	end

	-- sort world bosses by name (for now, maybe by worldboss id later)
	local wbsort = function(namea,nameb)
		local ta = RepeatableQuests[namea]
		local tb = RepeatableQuests[nameb]
		local ida = ta and ta.questIds[1] or 0
		local idb = tb and tb.questIds[1] or 0

		if ida == idb then
			return namea < nameb
		else
			return ida > idb
		end
	end
	tsort(WorldBosses, wbsort)
	spanStart = 1 + tooltip:GetColumnCount()
	for i, name in ipairs(WorldBosses) do
		MyAddColumn(strformat("WorldBosses.%s", name), self:GetInstanceAbbr(name)) -- abbr name actually, if abbr name doesn't exist we won't add
	end
	AddSpanner(spanStart, tooltip:GetColumnCount(), L["World Bosses"], profiledb.columnName.WorldBosses)


	-- first do the lockouts for newest raids, the ones that have RaidFinder versions as well which we use to set the order
	-- remove them from lockouts when done
	for _, parent in ipairs(RaidFinder) do
		local raid = Lockouts[parent.name]
		if raid then
			spanStart = 1 + tooltip:GetColumnCount()
			for _, type in ipairs(sortedInstanceTypes) do
				MyAddColumn(raid[type], self:GetDifficultyAbbr(type)) -- abbr of type, not type
			end
			for _, wing in ipairs(parent.wings) do
				-- REVIEW - we need to be checking the show.RaidFinder flag somehwere don't we? here?
				MyAddColumn(raid[wing.name], self:GetInstanceAbbr(wing.name)) -- abbr of wing.name
			end
			AddSpanner(spanStart, tooltip:GetColumnCount(), parent.name, self:GetInstanceAbbr(parent.name))
		end
		Lockouts[parent.name] = nil
	end

	-- then do the lockouts for legacy raids, the ones without RaidFinder versions (everything before Dragonsoul)
	-- remove them from lockouts when done
	for _, expac in ipairs(LegacyRaid) do
		spanStart = 1 + tooltip:GetColumnCount()
		for _, raid in ipairs(expac.raids) do
			local raidlocks = Lockouts[raid.name]
			-- print(raid.name, raidlocks)
			if raidlocks then
				local normal = raidlocks["NormalRaid"]
				if normal then
					MyAddColumn(normal, self:GetInstanceAbbr(raid.name)) -- abbr of type, not type
					raidlocks["NormalRaid"] = nil
				end
				local legacy = raidlocks["LegacyRaid"]
				if legacy then
					MyAddColumn(legacy, self:GetInstanceAbbr(raid.name)) -- abbr of type, not type
					raidlocks["LegacyRaid"] = nil
				end
			end
		end
		if tooltip:GetColumnCount() >= spanStart then
			AddSpanner(spanStart, tooltip:GetColumnCount(), expac.name, self:GetInstanceAbbr(expac.name))
		end
	end

	-- flip the lockouts table - now it will be type > name
	local LockTypes = {}
	for name, types in pairs(Lockouts) do
		for type, nameType in pairs(types) do
			tget(LockTypes,type)[name] = nameType
		end
	end
	for _, type in ipairs(sortedInstanceTypes) do
		spanStart = 1 + tooltip:GetColumnCount()
		local raids = LockTypes[type]
		if raids then
			for name, nameType in pairs(raids) do
				MyAddColumn(nameType, self:GetInstanceAbbr(name)) -- abbr of name
			end
		end
		AddSpanner(spanStart, tooltip:GetColumnCount(), type, self:GetDifficultyAbbr(type))
	end

	--==============================================================--
	-- END OF COLUMN CREATION
	--
	-- At this point, we have determined all the trackers that have
	-- data to be displayed. All of the columns needed for that data
	-- have been created and are in proper order.
	--==============================================================--


	--==============================================================--
	-- PREP ROWS
	--
	-- figure out who we need rows for and what the order will be
	--==============================================================--

	if not profiledb.showAllChars then
		for realmChar, charData in pairs(roster) do
			if not next(charData.resets) then
				-- print(strformat("%s dropped from roster for having no resets", realmChar))
				roster[realmChar] = nil
			end
		end
	end

	local sortedRoster = SortRosterSlice(roster)
	-- PrintSortedRosterSlice(sortedRoster)


	--==============================================================--
	-- FILL IN THE GRID
	--==============================================================--

	-- structure is { [iRealm] = { name, chars = { [iChar] = { name, level, color, resets = { name = data } } } } }
	-- eg count = sortedRoster[1].chars[1].resets.LootCoins.count

	for _, realm in ipairs(sortedRoster) do
		if profiledb.showRealmHeaderLines then
			tooltip:SetCell(tooltip:AddLine(), 1, realm.name, nil, "CENTER", tooltip:GetColumnCount())
		end

		for _, char in ipairs(realm.chars) do
			-- local charRealm = char.." - "..realm
			local lineName = char.name -- profile.showCharacterRealm and charRealm or char
			local lineFont = profiledb.useClassColors and char.color or nil --  RAID_CLASS_COLORS_FONTS[ self.db.global.charClass[charRealm] ] or nil
			local row = tooltip:AddLine("")
			tooltip:SetCell(row, 1, lineName, lineFont)

			-- init cells
			for col = tooltip:GetColumnCount(),2,-1 do
				-- TODO - if the char doesn't meet level requirement for column then don't init - leave a hole in the grid
				local column = Columns[col]
				column:Format(tooltip, row, col)
			end

			-- I think we could move this up into the previous loop by
			-- calling Format with char.resets[column.rest] after the col
			for name, data in pairs(char.resets) do
				-- print(name, Columns[name])
				local column = Columns[name]
				if column then
					column:Format(tooltip, row, column.col, data)
				else
					Ailo:Print("Error: missing column", name)
				end
			end
		end
	end
end


--=========================================================================--
--=========================================================================--
--
-- GenerateOptions - build options UI table
--
--=========================================================================--
--=========================================================================--

function Ailo.GenerateOptions()
	if not Ailo.options then
		local function newCounter()
			local i = 0
			return function()
					i = i + 1
					return i
				end
		end
		local count = newCounter()

		local getShow = function(info) return Ailo.db.profile.show[info[#info]] end
		local setShow = function(info, value) Ailo.db.profile.show[info[#info]] = value end

		Ailo.options = {
			name = "Ailo",
			type = 'group',
			handler = Ailo,
			args = {
				genconfig = {
					name = L["General Settings"],
					type = 'group',
					order = 1,
					get = function(info) return Ailo.db.profile[info[#info]] end,
					set = function(info, value) Ailo.db.profile[info[#info]] = value end,
					args = {
						savedraid = {
							name = L["Saved raid color"],
							desc = L["SAVED_RAID_DESC"],
							type = 'color',
							order = count(),
							get  = getColor,
							set  = setColor,
							hasAlpha = true,
						},
						freeraid = {
							name = L["Free raid color"],
							desc = L["FREE_RAID_DESC"],
							type = 'color',
							order = count(),
							get  = getColor,
							set  = setColor,
							hasAlpha = true,
						},
						useClassColors = {
							type = "toggle",
							order = count(),
							name = L["Color names by class"],
						},
						useCustomClassColors = {
							type = "toggle",
							order = count(),
							name = L["Use !ClassColors"],
							desc = L["Use !ClassColors addon for class colors used to color the names in the tooltip"],
							get = function(info) return Ailo.db.profile[info[#info]] end,
							set = function(info, value) 
								Ailo.db.profile[info[#info]] = value 
								Ailo:SetupClasscoloredFonts()
							end,
							disabled = function() return not Ailo.db.profile.useClassColors or not CUSTOM_CLASS_COLORS end,
						},
						showRealmHeaderLines  = {
							type = "toggle",
							order = count(),
							name = L["Show Realm Headers"],
							desc = L["SHOW_REALMLINES_DESC"],
						},
						showCharacterRealm = {
							name = L["Show character realms"],
							type = "toggle",
							order = count(),
						},
						minimapIcon = {
							type = "toggle",
							name = L["Show minimap button"],
							desc = L["Show the Ailo minimap button"],
							order = count(),
							get = function(info) return not Ailo.db.profile.minimapIcon.hide end,
							set = function(info, value)
								Ailo.db.profile.minimapIcon.hide = not value
								if value then LDBIcon:Show("Ailo") else LDBIcon:Hide("Ailo") end
							end,
						},
						showMessages = {
							type = "toggle",
							order = count(),
							name = L["Chatframe Messages"],
						},
						weeklyResetDay = {
							type = "select",
							order = count(),
							name = L["Weekly reset day"],
							desc = L["Day when weekly caps and lockouts are reset"],
							values = CALENDAR_WEEKDAY_NAMES,
							get = function(info) return Ailo.db.global.weeklyResetDay end,
							set = function(info, value) Ailo.db.global.weeklyResetDay = value end,
						},
						wipeDB = {
							type = "execute",
							name = L["Wipe Database"],
							order = count(),
							confirm = true,
							func = function() 
									Ailo:WipeDB()
									Ailo:InitCurrenciesTable() -- because we are caching totals and only write to db when they change
									Ailo:UpdatePlayer()
								end,
						},
						showAllChars  = {
							type = "toggle",
							order = count(),
							name = L["Show all chars"],
							desc = L["Regardles of any saved instances"],
						},                    
						showLeft = {
							type = "toggle",
							order = count(),
							name = L["Encounters Left"],
							desc = L["SHOW_ENCOUTERS_LEFT_DESC"],
						},
						minLevel = {
							type = "range",
							order = count(),
							min = 1;
							max = GetMaxLevelForExpansionLevel(GetExpansionLevel());
							step = 1,
							name = L["Minimum Level"],
							desc = L["Characters below this level will not be shown"],
							set = "LevelCheck",
						},

						-- headers and checkboxes for the trackers will be inserted here by code below
					},
				},
				columnName = {
					type = 'group',
					name = L["Column Names"],
					get = function(info) return Ailo.db.profile.columnName[info[#info]] end,
					set = function(info, value) Ailo.db.profile.columnName[info[#info]] = value end,
					args = {
						header = {
							type = "header",
							order = 1,
							name = L["Change the column names used in the tooltip"]
						},
						[WEEKDAY_SUNDAY] = {
							type = "input",
							order = 2,
							name = WEEKDAY_SUNDAY,
						},
						[WEEKDAY_MONDAY] = {
							type = "input",
							order = 3,
							name = WEEKDAY_MONDAY,
						},
						[WEEKDAY_TUESDAY] = {
							type = "input",
							order = 4,
							name = WEEKDAY_TUESDAY,
						},
						[WEEKDAY_WEDNESDAY] = {
							type = "input",
							order = 5,
							name = WEEKDAY_WEDNESDAY,
						},
						[WEEKDAY_THURSDAY] = {
							type = "input",
							order = 6,
							name = WEEKDAY_THURSDAY,
						},
						[WEEKDAY_FRIDAY] = {
							type = "input",
							order = 7,
							name = WEEKDAY_FRIDAY,
						},
						[WEEKDAY_SATURDAY] = {
							type = "input",
							order = 8,
							name = WEEKDAY_SATURDAY,
						},
						header2 = {
							type = "header",
							order = 9,
							name = ""
						},
						Dailies = {
							type = "input",
							name = L["Dailies"],
						},
						Weeklies = {
							type = "input",
							name = L["Weeklies"],
						},
						NZothAssaults = {
							type = "input",
							name = L["N'Zoth Assaults"],
						},
						Bounties = {
							type = "input",
							name = L["Bounties"],
						},
						WorldBosses = {
							type = "input",
							name = L["World Bosses"],
						},
						Total = {
							type = "input",
							name = L["Total"],
						},
						Weekly = {
							type = "input",
							name = L["Weekly"],
						},
						Conquest = {
							type = "input",
							name = L["Conquest"],
						},
						MythicPlus = {
							type = "input",
							name = L["Mythic Plus"],
						},
						GarrisonInvasion = {
							type = "input",
							name = L["Garrison Invasion"],
						},
						Renown = {
							type = "input",
							name = L["Renown"],
						},
						Callings = {
							type = "input",
							name = L["Callings"],
						},
						Knowledge = {
							type = "input",
							name = L["Profession Knowledge"],
						},
						PKTreatises = {
							type = "input",
							name = L["Draconic Treatises"],
						},
						PKTrainers = {
							type = "input",
							name = L["Profession Trainers"],
						},
						PKTraders = {
							type = "input",
							name = L["Faction Traders"],
						},
						PKOrders = {
							type = "input",
							name = L["Work Orders"],
						},
					},
				},
				instanceAbbr = { 
					type = 'group',
					name = L["Instance Abbreviations"],
					get = function(info) return Ailo.db.profile.instanceAbbr[info[#info]] end,
					set = function(info, value) Ailo.db.profile.instanceAbbr[info[#info]] = value end,
					args = {
						header = {
							type = "header",
							order = 1,
							name = L["Change the abbreviations used in the tooltip"]
						},
						[strNormal] = {
							type = "input",
							order = 2,
							name = strNormal,
							get = function(info) return Ailo.db.profile.difficultyAbbr[info[#info]] end,
							set = function(info, value) Ailo.db.profile.difficultyAbbr[info[#info]] = value end,
						},
						[strHeroic] = {
							type = "input",
							order = 3,
							name = strHeroic,
							get = function(info) return Ailo.db.profile.difficultyAbbr[info[#info]] end,
							set = function(info, value) Ailo.db.profile.difficultyAbbr[info[#info]] = value end,
						},
						[strMythic] = {
							type = "input",
							order = 4,
							name = strMythic,
							get = function(info) return Ailo.db.profile.difficultyAbbr[info[#info]] end,
							set = function(info, value) Ailo.db.profile.difficultyAbbr[info[#info]] = value end,
						},
						[strTimewalker] = {
							type = "input",
							order = 5,
							name = strTimewalker,
							get = function(info) return Ailo.db.profile.difficultyAbbr[info[#info]] end,
							set = function(info, value) Ailo.db.profile.difficultyAbbr[info[#info]] = value end,
						},
						[strChallenge] = {
							type = "input",
							order = 6,
							name = strChallenge,
							get = function(info) return Ailo.db.profile.difficultyAbbr[info[#info]] end,
							set = function(info, value) Ailo.db.profile.difficultyAbbr[info[#info]] = value end,
						},
						header2 = {
							type = "header",
							order = 7,
							name = ""
						},
					},
				},
			},
		}

		for instance, abbr in pairs(Ailo.db.profile.instanceAbbr) do
			Ailo.options.args.instanceAbbr.args[instance] = { type = "input", name = instance }
		end


		local op = Ailo.options.args.genconfig.args
		local AddCheckbox = function(header, tracker)
			if header == tracker.optionGroup then
				op[tracker.key] = {
					get = getShow,
					set = setShow,
					type = "toggle",
					order = count(),
					image = tracker.icon,
					name = tracker.optionName,
					desc = tracker.optionDesc,
				}
			end
		end

		for _, header in ipairs(sortedOptionGroups) do
			op["header"..header] = { type = "header", order = count(), name = optionGroupNames[header] }
			for _, group in ipairs(Trackers) do
				AddCheckbox(header, group)
				if group.children then
					for _, tracker in ipairs(group.children) do
						AddCheckbox(header, tracker)
					end
				end
			end

			for _, tracker in ipairs(SpecialTrackers) do
				AddCheckbox(header, tracker)
			end
		end

		Ailo.options.args.profile = LibStub("AceDBOptions-3.0"):GetOptionsTable(Ailo.db)
	end

	return Ailo.options
end


--=========================================================================--
--=========================================================================--
--
-- Debugging Functions
--
--=========================================================================--
--=========================================================================--

function Ailo:PrintGlobalRoster()
	local roster = self.db.global.roster
	local temp = {}
	for char,data in pairs(roster) do
		tinsert(temp, char)
	end
	tsort(temp)

	self:Print("globaldb roster")
	for i,char in ipairs(temp) do
		self:Print(i,char)
	end
	self:Print("")
end

function Ailo:PrintRosterSlice(roster)
	local temp = {}
	for realmChar,data in pairs(roster) do
		tinsert(temp, realmChar)
	end
	tsort(temp)

	self:Print("roster slice")
	for i,realmChar in ipairs(temp) do
		self:Print(i,realmChar, roster[realmChar].level, roster[realmChar].color)
	end
	self:Print("")
end

function Ailo:PrintSortedRosterSlice(roster)
	self:Print("sorted roster slice")
	for _,realm in ipairs(roster) do
		self:Print(realm.name)
		for _,char in ipairs(realm.chars) do
			self:Print("  ", char.name)
			for name,_ in pairs(char.resets) do
				self:Print("  ", "  ", name)
			end
		end
	end
	self:Print("")
end
